##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""The module to prepare the simulation from the concrete scenario."""

# import four elements of AbstractScenario
from ._sumo import SUMOPrep
from ._aimsun import AimsunPrep
from ._vissim import VissimPrep


class SimPrep:
    '''Prepare simulation from concrete scenario'''

    def __init__(self, **kwargs):
        self.SUMOSim = SUMOPrep(**kwargs)
        self.AimsunSim = AimsunPrep(**kwargs)
        self.VissimSim = VissimPrep(**kwargs)

    def create_sumo_sim(self,
                        ConcreteScn,
                        start_time: float = 3600 * 8,
                        end_time: float = 3600 * 10,
                        seed: list | int = 812,
                        step_length: float = 0.1):
        """Prepare SUMO documents for simulation.
        """

        # check seed type
        if isinstance(seed, int):
            seed = [seed]
        elif isinstance(seed, list):
            pass
        else:
            raise ValueError("  :seed must be an integer or a list of integers.")
        if len(seed) > 1:
            print("  :Multiple seeds are provided, the first one will be used.")
            seed = seed[0]

        self.SUMOSim.importNetwork(ConcreteScn)
        self.SUMOSim.importDemand(ConcreteScn,
                                  start_time,
                                  end_time,
                                  seed)
        self.SUMOSim.generateConfig(ConcreteScn,
                                    start_time,
                                    end_time,
                                    seed,
                                    step_length)
        self.SUMOSim.importSignal(ConcreteScn)
        # print("  :SUMO simulation is prepared.")

    def create_aimsun_sim(self,
                          ConcreteScn,
                          start_time: float = 3600 * 8,
                          end_time: float = 3600 * 10,
                          seed: list | int = 812,
                          step_length: float = 0.1):
        """Prepare Aimsun documents for simulation."""
        self.AimsunSim.importDemand(ConcreteScn, start_time, end_time)

    def create_vissim_sim(self,
                          ConcreteScn,
                          start_time: float = 3600 * 8,
                          end_time: float = 3600 * 10,
                          seed: list | int = 812,
                          step_length: float = 0.1):
        """Prepare VISSIM documents for simulation."""

        print("  :VISSIM simulation preparation is not implemented yet.")
        self.VissimSim.verbose = True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

import urllib.request


def download_single_file_from_web(url: str, dest_filename: str, chunk_size=1024) -> bool:
    """Downloads a large file from a URL in chunks and saves it to the specified destination.

    Args:
        url (str): The URL of the file to download.
        dest_filename (str): filename or path to the filename to save the downloaded file.
        chunk_size (int): Size of each chunk to read in bytes (default: 1024).

    Returns:
        bool: True if the download is successful, False otherwise.
    """
    try:
        with urllib.request.urlopen(url) as response, open(dest_filename, 'wb') as out_file:
            total_size = int(response.getheader('Content-Length', 0))

            if total_size == 0:
                print("  :An error occurred: File size is 0.")
                return False

            downloaded = 0

            print(f"  :Starting download: {url}")
            # print(f"  :Total size: {total_size / (1024 * 1024):.2f} MB")

            while chunk := response.read(chunk_size):
                out_file.write(chunk)
                downloaded += len(chunk)
                print(f"\r  :Downloaded: {downloaded / (1024 * 1024):.2f} / {total_size / (1024 * 1024):.2f}  MB",
                      end="")

    except Exception as e:
        print(f"  :An error occurred: {e}")
        return False
    return True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

import os
from pathlib import Path
import pyufunc as pf


def check_abstract_inputs(input_dir: str) -> bool:
    """ Check the input directory for the abstract scenario."""

    # check If Control folder is empty
    path_control = pf.path2linux(Path(input_dir) / "Control")
    if not os.path.exists(path_control):
        raise Exception(f"  :Error: Control folder does not exist: {path_control}")

    if not pf.get_filenames_by_ext(path_control, file_ext="csv"):
        raise Exception(f"  :Error: Control folder is empty: {path_control},"
                        "Please include Synchro UTDF file (signal) inside Control folder"
                        " and add the control file name to the input configuration file.")

    # check If Traffic folder is empty
    path_traffic = pf.path2linux(Path(input_dir) / "Traffic")
    if not os.path.exists(path_traffic):
        raise Exception(f"  :Error: Traffic folder does not exist: {path_traffic}")

    if not pf.get_filenames_by_ext(path_traffic, file_ext="*"):
        raise Exception(f"  :Error: Traffic folder is empty: {path_traffic},"
                        "Please include turn movement file for each intersection inside Traffic folder"
                        " and add the file names to the MatchupTable.xlsx "
                        "(You will notice the generated MatchupTable.xlsx inside your input folder).")

    # check if the MatchupTable.xlsx exists
    path_matchup = pf.path2linux(Path(input_dir) / "MatchupTable.xlsx")
    if not os.path.exists(path_matchup):
        raise Exception(f"  :Error: Matchup table does not exist: {path_matchup}"
                        "Please generate the Matchup table by run generate_inputs()")
    return True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
import xml.etree.ElementTree as ET
from xml.dom.minidom import parseString
import shutil
import os
from pathlib import Path
import pyufunc as pf
import copy

from realtwin.func_lib._f_calibration.algo_sumo.cali_turn_inflow import TurnInflowCali
from realtwin.func_lib._f_calibration.algo_sumo.cali_behavior import BehaviorCali
from realtwin.func_lib._f_calibration.algo_sumo.util_cali_turn_inflow import (read_MatchupTable,
                                                                              generate_turn_demand_cali,
                                                                              generate_inflow,
                                                                              generate_turn_summary)
from realtwin.func_lib._f_calibration.algo_sumo.util_cali_behavior import auto_select_two_routes


# for the beta version
def cali_sumo(*, sel_algo: dict = None, input_config: dict = None, verbose: bool = True, **kwargs) -> bool:
    """Run SUMO calibration based on the selected algorithm and input configuration.

    Args:
        sel_algo (dict): the dictionary of selected algorithm for turn_inflow and behavior. Defaults to None.
        input_config (dict): the dictionary contain configurations from input yaml file. Defaults to None.
        verbose (bool): print out processing message. Defaults to True.

    Raises:
        ValueError: if algo_config is not a dict with two levels with keys of 'ga', 'sa', and 'ts'
        ValueError: if sel_algo is not a dict with keys of 'turn_inflow' and 'behavior'

    Returns:
        bool: True if calibration is successful, False otherwise.
    """

    # Test-driven Development: check selected algorithm from input
    if sel_algo is None:  # use default algorithm if not provided
        sel_algo = {"turn_inflow": "ga", "behavior": "ga"}

    if not isinstance(sel_algo, dict):
        print("  :Error:parameter sel_algo must be a dict with"
              " keys of 'turn_inflow' and 'behavior', using"
              " genetic algorithm as default values.")
        sel_algo = {"turn_inflow": "ga", "behavior": "ga"}

    # Prepare scenario_config and algo_config from input_config
    scenario_config_turn_inflow = prepare_scenario_config_turn_inflow(input_config)

    # Prepare Algorithm configure: e.g. {"ga": {}, "sa": {}, "ts": {}}
    algo_config_turn_inflow = input_config["Calibration"]["turn_inflow"]
    algo_config_turn_inflow["ga_config"] = input_config["Calibration"]["ga_config"]
    algo_config_turn_inflow["sa_config"] = input_config["Calibration"]["sa_config"]
    algo_config_turn_inflow["ts_config"] = input_config["Calibration"]["ts_config"]

    if "update_turn_flow_algo" in kwargs:
        algo_config_turn_inflow["ga_config"] = algo_config_turn_inflow["ga_config"].update(
            kwargs["update_turn_flow_algo"].get("ga_config", {}))
        algo_config_turn_inflow["sa_config"] = algo_config_turn_inflow["sa_config"].update(
            kwargs["update_turn_flow_algo"].get("sa_config", {}))
        algo_config_turn_inflow["ts_config"] = algo_config_turn_inflow["ts_config"].update(
            kwargs["update_turn_flow_algo"].get("ts_config", {}))

    algo_config_behavior = input_config["Calibration"]["behavior"]
    algo_config_behavior["ga_config"] = input_config["Calibration"]["ga_config"]
    algo_config_behavior["sa_config"] = input_config["Calibration"]["sa_config"]
    algo_config_behavior["ts_config"] = input_config["Calibration"]["ts_config"]

    if "update_behavior_algo" in kwargs:
        algo_config_behavior["ga_config"] = algo_config_behavior["ga_config"].update(
            kwargs["update_behavior_algo"].get("ga_config", {}))
        algo_config_behavior["sa_config"] = algo_config_behavior["sa_config"].update(
            kwargs["update_behavior_algo"].get("sa_config", {}))
        algo_config_behavior["ts_config"] = algo_config_behavior["ts_config"].update(
            kwargs["update_behavior_algo"].get("ts_config", {}))

    # run calibration based on the selected algorithm: optimize turn and inflow
    print("\n  :Optimize Turn and Inflow...")
    turn_inflow = TurnInflowCali(scenario_config_turn_inflow, algo_config_turn_inflow, verbose=verbose)

    match sel_algo["turn_inflow"]:
        case "ga":
            g_best, model = turn_inflow.run_GA()
            path_model_result = "turn_inflow_ga_result"
        case "sa":
            g_best, model = turn_inflow.run_SA()
            path_model_result = "turn_inflow_sa_result"
        case "ts":
            g_best, model = turn_inflow.run_TS()
            path_model_result = "turn_inflow_ts_result"
        case _:
            print(f"  :Error: unsupported algorithm {sel_algo['turn_inflow']}, using genetic algorithm as default.")
            g_best, model = turn_inflow.run_GA()
            path_model_result = "turn_inflow_ga_result"

    turn_inflow.run_vis(path_model_result, model)
    # clean up the temporary files generated during turn and inflow calibration
    turn_inflow._clean_up()

    print("\n  :Optimize Behavior parameters based on the optimized turn and inflow...")
    scenario_config_behavior = prepare_scenario_config_behavior(input_config)
    if "sel_behavior_routes" in kwargs:
        scenario_config_behavior["sel_behavior_routes"] = kwargs["sel_behavior_routes"]
    else:
        # automatically select two routes from network
        dir_behavior = scenario_config_behavior["dir_behavior"]
        network_name = input_config.get("Network").get("NetworkName")
        path_net = Path(dir_behavior) / f"{network_name}.net.xml"
        path_route = Path(dir_behavior) / f"{network_name}.rou.xml"
        path_report = Path(dir_behavior) / "selected_routes_travel_time_map.html"
        google_api = ""
        routes_list, time_list, edge_list = auto_select_two_routes(path_route, path_net,
                                                                   api_key=google_api, path_report=path_report)
        print(f"  : selected routes: {routes_list}")
        print(f"  : selected travel time: {time_list}")
        print(f"  : selected edge list: {edge_list}")
        sel_route_dict = {}
        route_id = 1
        for route_name, travel_time, edge_id_list in zip(routes_list, time_list, edge_list):
            sel_route_dict[f"route_{route_id}"] = {"time": travel_time,
                                                   "edge_list": edge_id_list,
                                                   "route_list": route_name}
            route_id += 1
        scenario_config_behavior["sel_behavior_routes"] = sel_route_dict
    print(f"  \n:Selected behavior routes: {scenario_config_behavior['sel_behavior_routes']}\n")
    behavior = BehaviorCali(scenario_config_behavior, algo_config_behavior, verbose=verbose)

    match sel_algo["behavior"]:
        case "ga":
            g_best, model = behavior.run_GA()
            path_model_result = "behavior_ga_result"
        case "sa":
            g_best, model = behavior.run_SA()
            path_model_result = "behavior_sa_result"
        case "ts":
            g_best, model = behavior.run_TS()
            path_model_result = "behavior_ts_result"
        case _:
            print(f"  :Error: unsupported algorithm {sel_algo['behavior']}, using genetic algorithm as default.")
            g_best, model = behavior.run_GA()
            path_model_result = "behavior_ga_result"

    behavior.run_vis(path_model_result, model)
    return True


def prepare_scenario_config_turn_inflow(input_config: dict) -> dict:
    """Prepare scenario_config from input_config"""

    scenario_config_dict = input_config.get("Calibration").get("scenario_config")

    # # add input_dir to scenario_config from generated SUMO dir(scenario generation)
    generated_sumo_dir = pf.path2linux(Path(input_config["output_dir"]) / "SUMO")
    # generated_sumo_dir = pf.path2linux(Path(__file__).parents[3] / "datasets/input_dir_dummy/")
    # print(f"  :use dummy input: {generated_sumo_dir} for calibration in beta version")

    # create turn_inflow directory under generated_sumo_dir
    turn_inflow_dir = pf.path2linux(Path(generated_sumo_dir) / "turn_inflow")
    os.makedirs(turn_inflow_dir, exist_ok=True)
    turn_inflow_route_dir = pf.path2linux(Path(generated_sumo_dir) / "turn_inflow" / "route")
    os.makedirs(turn_inflow_route_dir, exist_ok=True)

    # add input_dir as turn_inflow_dir
    scenario_config_dict["dir_turn_inflow"] = turn_inflow_dir

    # copy net.xml to turn_inflow directory
    network_name = input_config.get("Network").get("NetworkName")
    path_net_sumo = pf.path2linux(Path(generated_sumo_dir) / f"{network_name}.net.xml")
    shutil.copy(path_net_sumo, turn_inflow_dir)
    # shutil.copy(path_net_sumo, turn_inflow_route_dir)

    # create Edge.add.xml in turn_inflow directory
    path_edge_add = pf.path2linux(Path(turn_inflow_dir) / "Edge.add.xml")
    generate_edge_add_xml(path_edge_add)

    # create .cfg file in turn_inflow directory
    path_sumocfg = pf.path2linux(Path(turn_inflow_dir) / f"{network_name}.sumocfg")
    seed = scenario_config_dict.get("calibration_seed", 812)
    sim_start_time = scenario_config_dict.get("sim_start_time", 3600 * 8)
    sim_end_time = scenario_config_dict.get("sim_end_time", 3600 * 10)
    calibration_time_step = scenario_config_dict.get("calibration_time_step", 1)
    generate_sumocfg_xml(path_sumocfg, network_name, seed, sim_start_time, sim_end_time, calibration_time_step)

    # create turn and inflow and summary df
    path_matchup_table = pf.path2linux(Path(input_config["input_dir"]) / "MatchupTable.xlsx")
    traffic_dir = pf.path2linux(Path(input_config["input_dir"]) / "Traffic")
    path_net_turn_inflow = pf.path2linux(Path(turn_inflow_dir) / f"{network_name}.net.xml")
    MatchupTable_UserInput = read_MatchupTable(path_matchup_table=path_matchup_table)
    TurnDf, IDRef = generate_turn_demand_cali(path_matchup_table=path_matchup_table, traffic_dir=traffic_dir)

    InflowDf_Calibration, InflowEdgeToCalibrate, N_InflowVariable = generate_inflow(path_net_turn_inflow,
                                                                                    MatchupTable_UserInput,
                                                                                    TurnDf,
                                                                                    IDRef)

    (TurnToCalibrate, TurnDf_Calibration,
     RealSummary_Calibration,
     N_Variable, N_TurnVariable) = generate_turn_summary(TurnDf,
                                                         MatchupTable_UserInput,
                                                         N_InflowVariable)

    scenario_config_dict["TurnToCalibrate"] = TurnToCalibrate
    scenario_config_dict["TurnDf_Calibration"] = TurnDf_Calibration
    scenario_config_dict["InflowDf_Calibration"] = InflowDf_Calibration
    scenario_config_dict["InflowEdgeToCalibrate"] = InflowEdgeToCalibrate
    scenario_config_dict["RealSummary_Calibration"] = RealSummary_Calibration
    scenario_config_dict["N_InflowVariable"] = N_InflowVariable
    scenario_config_dict["N_Variable"] = N_Variable
    scenario_config_dict["N_TurnVariable"] = N_TurnVariable

    # add network name to scenario_config and sim_name
    scenario_config_dict["network_name"] = network_name
    scenario_config_dict["sim_name"] = f"{network_name}.sumocfg"
    scenario_config_dict["path_net"] = pf.path2linux(Path(turn_inflow_dir) / f"{network_name}.net.xml")

    return scenario_config_dict


def prepare_scenario_config_behavior(input_config: dict) -> dict:

    scenario_config_behavior = input_config.get("Calibration").get("scenario_config")
    network_name = input_config.get("Network").get("NetworkName")

    # # add input_dir to scenario_config from generated SUMO dir(scenario generation)
    generated_sumo_dir = pf.path2linux(Path(input_config["output_dir"]) / "SUMO")

    # create behavior directory under generated_sumo_dir
    behavior_dir = pf.path2linux(Path(generated_sumo_dir) / "behavior")
    os.makedirs(behavior_dir, exist_ok=True)
    scenario_config_behavior["dir_behavior"] = behavior_dir

    # copy files from turn_inflow directory to behavior directory
    turn_inflow_dir = pf.path2linux(Path(generated_sumo_dir) / "turn_inflow")
    file_sim = Path(turn_inflow_dir) / f"{network_name}.sumocfg"
    file_net = Path(turn_inflow_dir) / f"{network_name}.net.xml"
    file_edge_add = Path(turn_inflow_dir) / "Edge.add.xml"
    file_route = Path(turn_inflow_dir) / f"{network_name}.rou.xml"
    file_turn = Path(turn_inflow_dir) / f"{network_name}.turn.xml"
    file_inflow = Path(turn_inflow_dir) / f"{network_name}.flow.xml"
    file_EdgeData = Path(turn_inflow_dir) / "EdgeData.xml"
    shutil.copy(file_sim, behavior_dir)
    shutil.copy(file_net, behavior_dir)
    shutil.copy(file_edge_add, behavior_dir)
    shutil.copy(file_route, behavior_dir)
    shutil.copy(file_turn, behavior_dir)
    shutil.copy(file_inflow, behavior_dir)
    shutil.copy(file_EdgeData, behavior_dir)

    scenario_config_behavior["path_turn"] = f"{network_name}.turn.xml"
    scenario_config_behavior["path_inflow"] = f"{network_name}.flow.xml"
    return scenario_config_behavior


def generate_edge_add_xml(path_edge_add: str) -> bool:
    """Generate Edge.add.xml file in the input directory.

    Args:
        input_config (dict): the dictionary contain configurations from input yaml file.
    """
    # create Edge.add.xml in turn_inflow directory
    additional = ET.Element("additional")
    edgeData = ET.SubElement(additional, "edgeData")
    edgeData.set("id", "1")
    edgeData.set("file", "EdgeData.xml")
    tree = ET.ElementTree(additional)
    tree.write(path_edge_add, encoding="utf-8", xml_declaration=True)
    return True


def generate_sumocfg_xml(path_sumocfg: str, network_name: str, seed: int,
                         sim_start_time: int, sim_end_time: int, calibration_time_step: int) -> bool:
    # create .cfg file in turn_inflow directory
    # Create XML root
    root = ET.Element('configuration')
    root.set('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    root.set('xsi:noNamespaceSchemaLocation', 'http://sumo.dlr.de/xsd/duarouterConfiguration.xsd')

    # Random seed
    random = ET.SubElement(root, 'random')
    ET.SubElement(random, 'seed', {'value': str(seed)})

    # Input files
    input_elem = ET.SubElement(root, 'input')
    ET.SubElement(input_elem, 'net-file', {'value': f'{network_name}.net.xml'})
    ET.SubElement(input_elem, 'route-files', {'value': f'{network_name}.rou.xml'})
    ET.SubElement(input_elem, 'additional-files', {'value': 'Edge.add.xml'})

    # Output (empty section placeholder)
    ET.SubElement(root, 'output')

    # Time setup
    time = ET.SubElement(root, 'time')
    ET.SubElement(time, 'begin', {'value': str(sim_start_time)})
    ET.SubElement(time, 'end', {'value': str(sim_end_time)})
    ET.SubElement(time, 'step-length', {'value': str(calibration_time_step)})

    # GUI options
    gui_only = ET.SubElement(root, 'gui_only')
    ET.SubElement(gui_only, 'start', {'value': 't'})

    # Report options
    report = ET.SubElement(root, 'report')
    ET.SubElement(report, 'no-warnings', {'value': 'true'})
    ET.SubElement(report, 'no-step-log', {'value': 'true'})

    # Pretty print
    rough_string = ET.tostring(root, 'utf-8')
    reparsed = parseString(rough_string)
    xml_string = reparsed.toprettyxml(indent="    ")

    # Write to file
    with open(path_sumocfg, 'w') as file:
        file.write(xml_string)
    return True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

import re


def get_bbox_from_vertices(vertices: str | list) -> tuple:
    """get the bounding box from the vertices string

    Args:
        vertices (str): the vertices of the network in string format
            "(lon, lat),(lon, lat),..."

    Notes:
        The vertices format can be found in configuration file

    Returns:
        tuple: the bounding box of the network: (min_lon, min_lat, max_lon, max_lat)
    """
    if isinstance(vertices, str):
        # Regular expression to extract the coordinate pairs
        pattern = r"\((-?\d+\.\d+),\s*(-?\d+\.\d+)\)"
        matches = re.findall(pattern, vertices)

        lon_lst = [float(match[0]) for match in matches]
        lat_lst = [float(match[1]) for match in matches]
    elif isinstance(vertices, (list, tuple)):
        # Check if the list contains tuples
        if all(isinstance(item, (list, tuple)) and len(item) == 2 for item in vertices):
            lon_lst = [float(item[0]) for item in vertices]
            lat_lst = [float(item[1]) for item in vertices]
        else:
            raise ValueError(
                "Invalid format: List must contain list/tuple of [lon, lat].")
    else:
        raise ValueError("Invalid format: vertices must be a string or list.")

    return (min(lon_lst), min(lat_lst), max(lon_lst), max(lat_lst))



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
"""Control of module imports for the RealTwin function library."""

import os
import zipfile
import pyufunc as pf

from realtwin.func_lib._a_install_simulator.check_sim_env import is_sumo_installed
from realtwin.util_lib.download_file_from_web import download_single_file_from_web


def install_sumo(sel_dir: list = None,
                 strict_sumo_version: str = "1.21.0",
                 verbose: bool = True,
                 **kwargs) -> bool:
    """Install the SUMO simulator.

    Args:
        sel_dir (list): A list of directories to search for the SUMO executable. Defaults to None.
        strict_sumo_version (bool): If True, check and install the exact version of SUMO. Default is 1.21.0
        verbose (bool): If True, print the installation process. Default is True.
        kwargs: Additional keyword arguments.

    Returns:
        bool: True if the SUMO is installed successfully, False otherwise
    """

    # check sel_dir is a list
    if not isinstance(sel_dir, (list, type(None))):
        raise ValueError("sel_dir should be a list.")

    # Check if SUMO is already installed
    version_lst = is_sumo_installed(sel_dir=sel_dir, verbose=verbose)
    if version_lst:
        # Check if the exact version of SUMO is installed
        if strict_sumo_version is None or strict_sumo_version in version_lst:
            print(f"  :SUMO is already installed, available versions: {version_lst}")
            return True

        print(f"\n  :Installing strict_sumo_version SUMO version {strict_sumo_version} "
              f"(Available versions: {version_lst})...")

    # If SUMO not installed,
    # Or strict_sumo_version is True and the version is not installed
    # Install the SUMO
    if pf.is_windows():
        return install_sumo_windows(strict_sumo_version, verbose=verbose)

    if pf.is_linux():
        print("  :Error: Linux is not supported yet.")
        return False

    if pf.is_mac():
        print("  :Error: MacOS is not supported yet.")
        return False

    print("  :Error: Unsupported operating system.")
    return False


def install_sumo_windows(sumo_version: str = "1.21.0", verbose: bool = True) -> bool:
    """Install SUMO onto the windows system.

    Args:
        sumo_version (str): The version of SUMO to be installed. Default is "1.21.0".
        verbose (bool): If True, print the installation process. Default is True.

    Returns:
        bool: True if the SUMO is installed successfully, False otherwise
    """

    print(f"  :Installing SUMO {sumo_version} for Windows...")

    # Download SUMO from the official website
    sumo_release_url = "https://sumo.dlr.de/releases/"
    sumo_version_win = f"sumo-win64-{sumo_version}.zip"

    download_path = pf.path2linux(os.path.join(os.getcwd(), "sumo.zip"))
    extract_path = pf.path2linux(os.path.join(os.getcwd(), "SUMO"))

    # Download the SUMO zip file from the official website
    sumo_zip_url = f"{sumo_release_url}{sumo_version}/{sumo_version_win}"
    id_download = download_single_file_from_web(sumo_zip_url, download_path)

    if not id_download:
        return False

    # Extract the SUMO zip file
    if verbose:
        print(f"  :Extracting SUMO {sumo_version} for Windows at: {extract_path}...")

    with zipfile.ZipFile(download_path, 'r') as zip_ref:
        zip_ref.extractall(extract_path)

    # Clean up the downloaded zip file
    os.remove(download_path)

    # check if SUMO bin folder exists
    sumo_bin_path = os.path.join(extract_path, f"sumo-{sumo_version}", "bin")
    sumo_bin_path = pf.path2linux(sumo_bin_path)
    if not os.path.exists(sumo_bin_path):
        print(f"  :Error: bin folder not found in extracted SUMO directory: {sumo_bin_path}")
        return False

    # Add the SUMO bin folder to the system PATH
    if sumo_bin_path not in os.environ['PATH']:

        os.environ["PATH"] += os.pathsep + sumo_bin_path[0].upper() + sumo_bin_path[1:]
        os.environ["PATH"] += os.pathsep + sumo_bin_path[0].lower() + sumo_bin_path[1:]

        # add_path = subprocess.run(["setx", "PATH", f"%PATH%;{sumo_bin_path}"], shell=True, check=True)
        # if add_path.returncode == 0:
        #     print("  :SUMO is installed successfully.")
        # else:
        #     print("  :Error: Failed to add SUMO bin folder to system PATH.")

    return True


def install_sumo_linux(sumo_version: str = "1.21.0", verbose: bool = True) -> bool:
    """Install SUMO onto the linux system.

    Args:
        sumo_version (str): The version of SUMO to be installed. Default is "1.21.0".
        verbose (bool): If True, print the installation process. Default is True.

    Note:
        The installation of SUMO on Linux is not supported yet.
        upcoming feature.
    Returns:
        bool: True if the SUMO is installed successfully on Linux, False otherwise
    """
    print(f"  :Installing SUMO {sumo_version} for Linux system...")
    print("verbose", verbose)
    print("  :Warning: Install SUMO on Linux is not supported yet.")

    return False


def install_sumo_macos(sumo_version: str = "1.21.0", verbose: bool = True) -> bool:
    """Install SUMO onto the mac system.

    Args:
        sumo_version (str): The version of SUMO to be installed. Default is "1.21.0".
        verbose (bool): If True, print the installation process. Default is True.

    Note:
        The installation of SUMO on Mac is not supported yet.
        upcoming feature.

    Returns:
        bool: True if the SUMO is installed successfully on Mac, False otherwise
    """
    print(f"  :Installing SUMO {sumo_version} for Mac system...")
    print("verbose", verbose)
    print("  :Warning: Install SUMO on Mac is not supported yet.")

    return False



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host Application element of Abstract scenario
'''


class Application():
    '''The application class to host the Application element of Abstract scenario
    '''
    def __init__(self):
        pass

    def isEmpty(self):
        """Check if the Application element is empty"""
        pass

    def setValue(self, UserInputPath):
        """Set the value of the Application element from the user input file"""
        # self.CAVparameter == pd.read_csv(UserInputPath)
        pass



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host demand element of Concrete scenario
'''


class TrafficControl:
    '''The demand class to host the demand element of Concrete scenario
    '''
    def __init__(self):
        self.Signal = {}

    def is_empty(self):
        """Check if the TrafficControl object is empty."""
        pass

    def generate_control(self, AbsScn):
        """Generate control data from the abstract scenario."""
        # =================================
        # load class from AbstractScenario
        # =================================
        # SignalDict = AbsScn.dataObjDict['Control']['Signal'].Signal
        # # Read this for now
        # IDRef = pd.read_csv('Synchro_lookuptable.csv')
        # MergedDf3 = pd.merge(SignalDict ,IDRef, on=['INTID'], how='left')

        pass



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""The real-twin developed by ORNL Applied Research and Mobility System (ARMS) group"""

import os
import shutil
from pathlib import Path
import time
import pyufunc as pf
from rich.console import Console
from rich import print as rprint
# environment setup
from realtwin.util_lib.create_venv import venv_create, venv_delete
from realtwin.func_lib._a_install_simulator.inst_sumo import install_sumo

# input data loading
from realtwin.func_lib._b_load_inputs.loader_config import load_input_configs
from realtwin.util_lib.mapping_SUMO_OpenDrive_ID import parse_SUMO_ID, parse_SUMO_to_OpenDrive, parse_SUMO_TLS_ID

# scenario generation
# from realtwin.util_lib.download_elevation_tif import download_elevation_tif_by_bbox
from realtwin.util_lib.check_abstract_scenario_inputs import check_abstract_inputs
from realtwin.func_lib._c_abstract_scenario._abstractScenario import AbstractScenario
from realtwin.func_lib._c_abstract_scenario.rt_matchup_table_generation import generate_matchup_table
from realtwin.func_lib._c_abstract_scenario.rt_matchup_table_generation import format_junction_bearing
from realtwin.func_lib._c_abstract_scenario.rt_demand_generation import generate_turn_demand, update_matchup_table

from realtwin.func_lib._d_concrete_scenario._concreteScenario import ConcreteScenario

# simulation
from realtwin.func_lib._e_simulation._generate_simulation import SimPrep

# calibration
from realtwin.func_lib._f_calibration.calibration_sumo import cali_sumo
from realtwin.data_lib.config_data_lib import sel_behavior_routes as sel_behavior_routes_demo

console = Console()
# info: dim cyan, warning: magenta, danger: bold red


class RealTwin:
    """The real-twin developed by ORNL Applied Research and Mobility System (ARMS) group that
    enables the simulation of twin-structured cities.
    """

    def __init__(self, input_config_file: str = "", **kwargs):
        """Initialize the REALTWIN object.

        Args:
            input_config_file (str): The directory containing the input files.
            kwargs: Additional keyword arguments. Will be used in the future.
        """

        # initialize the input directory
        if not input_config_file:
            raise Exception(
                "\n  :Input configuration file is not provided."
                "\n  :RealTwin requires a configuration file to be provided.")

        self.input_config = load_input_configs(input_config_file)

        # add venv_create and delete as object methods
        self.venv_create = venv_create
        self.venv_delete = venv_delete
        self._venv_name = "venv_rt"
        self._proj_dir = os.getcwd()  # get current working directory

        # extract data from kwargs
        self.verbose = kwargs.get("verbose", False)

        # whether to stop the program to let user confirm input
        self._input_confirm = kwargs.get("input_confirm", True)

    def env_setup(self,
                  *,
                  sel_sim: list = None,
                  sel_dir: list = None,
                  strict_sumo_version: str = None,
                  strict_vissim_version: str = None,
                  strict_aimsun_version: str = None,
                  **kwargs) -> str:
        """Check and set up the environment for the simulation

        Args:
            sel_sim (list): select simulator to be set up. Default is None.
                Currently available options are ["SUMO", "VISSIM", "AIMSUN"].
            sel_dir (list): A list of directories to search for the executables. Defaults to None.
            strict_sumo_version (str): Whether to strictly check the version is installed.
                if specified, will check and install the version. Default is None.
            strict_vissim_version (str): Whether to strictly check the version is installed.
                if specified, will check and install the version. Default is False.
            strict_aimsun_version (str): Whether to strictly check the version is installed.
                if specified, will check and install the version. Default is False.
            kwargs: Additional keyword arguments.

        Examples:
            >>> import realtwin as rt
            >>> twin = rt.REALTWIN(input_config_file="config.yaml", verbose=True)

            check simulator is installed or not, default to SUMO, optional: VISSIM, AIMSUN
            >>> twin.env_setup(sel_sim=["SUMO"])

            add additional directories to search for the executables
            >>> additional_dir = [r"path-to-your-local-installed-sumo-bin"]
            >>> twin.env_setup(sel_sim=["SUMO"], sel_dir=additional_dir)

            strict version check: will install the required version if not found
            >>> twin.env_setup(sel_sim=["SUMO"], sumo_version="1.21.0", strict_sumo_version=True)

            or with additional directories
            >>> twin.env_setup(sel_sim=["SUMO"], sel_dir=additional_dir,
            >>>                sumo_version="1.21.0", strict_sumo_version=True)

        Returns:
            str: The selected simulator installation status.
        """

        # 0. Check if the sim_env is selected,
        #    default to SUMO, case insensitive and add self.sel_sim as a class attribute
        sel_sim = [sim.lower() for sim in sel_sim] if sel_sim else ["sumo"]

        # 1. Check simulator installation - mapping function
        simulator_installation = {
            "sumo": install_sumo,
            "vissim": None,
            "aimsun": None,
        }

        # 2. check if the simulator is installed, if not, install it
        console.print("\n[bold green]Check / Install the selected simulators:")

        kwargs['sel_dir'] = sel_dir
        kwargs['strict_sumo_version'] = strict_sumo_version
        kwargs['strict_vissim_version'] = strict_vissim_version
        kwargs['strict_aimsun_version'] = strict_aimsun_version
        kwargs['verbose'] = self.verbose

        invalid_sim = []
        for simulator in sel_sim:
            try:
                sim_status = simulator_installation.get(simulator)(**kwargs)
                if not sim_status:
                    invalid_sim.append(simulator)
            except Exception:
                invalid_sim.append(simulator)
                rprint(f"  :[bold magenta]Could not install {simulator} (strict version) on your operation system", end="")

        sel_sim_ = list(set(sel_sim) - set(invalid_sim))

        if not sel_sim_:
            raise Exception("  :Error: No simulator is available (strict version). Please select available version(s).")
        self.sel_sim = sel_sim_

        return f"  :Info: Selected simulators: {sel_sim_} are installed successfully."

    def generate_inputs(self, *, incl_sumo_net: str = None) -> str:
        """ Generate user inputs, such as MatchUp table, Control and Traffic data

        Args:
            incl_sumo_net (str): The path to the updated SUMO network file (.net.xml) provided by the user.
                If provided, the OpenDrive network will be generated based on this SUMO network.
                If not provided, the OpenDrive network will be generated based on the vertices from the config file.

        See Also:
            - How to create configuration file
            - How to create/update MatchUp table
            - How to create/prepare Control and Traffic data
            - How to download elevation tif data from network BBOX

        Returns:
            str: The status of the input generation.
        """

        with console.status("[bold cyan]Generating inputs...", spinner="dots"):
            console.print("\n[bold green]Check / Create input files and folders for user:")

            # remove output directory if it exists
            if Path(self.input_config.get("output_dir")).exists():
                shutil.rmtree(self.input_config.get("output_dir"))

            path_input = pf.path2linux(Path(self.input_config.get("input_dir")))

            # check if Control folder exists in the input directory
            path_control = pf.path2linux(Path(path_input) / "Control")
            if not os.path.exists(path_control):
                os.makedirs(path_control)

            # check if the Control folder is empty
            elif not os.listdir(path_control):
                console.print(f"[dim cyan]Control folder is empty: {path_control}.")

            console.print(f"  [dim cyan]:Control folder exists: {path_control}.[/dim cyan]\n"
                          "  :NOTICE: [bold red]Please include Synchro UTDF file (signal) inside Control folder\n")

            # check if Traffic folder exists in the input directory
            path_traffic = pf.path2linux(Path(path_input) / "Traffic")
            if not os.path.exists(path_traffic):
                os.makedirs(path_traffic)

            # check if the Traffic folder is empty
            elif not os.listdir(path_traffic):
                console.print(f"  [magenta]:Traffic folder is empty: {path_traffic}.")

            console.print(f"  [dim cyan]:Traffic folder exists: {path_traffic}.[/dim cyan]\n"
                          "  :NOTICE: [bold red]Please include turn movement file for each intersection "
                          "inside Traffic folder and add the file names to the MatchupTable.xlsx "
                          "(You will notice the generated MatchupTable.xlsx inside your input folder)."
                          " For how to fill the MatchupTable.xlsx, please refer to official documentation\n",
                          soft_wrap=True, no_wrap=False)

            # check if SUMO net file generated (in OpenDrive folder), if not, create the net.
            net_name = self.input_config["Network"]["NetworkName"]
            path_sumo_net = pf.path2linux(Path(self.input_config.get("output_dir")) / f"OpenDrive/{net_name}.net.xml")

            # generate abstract scenario if sumo net file does not exist
            self.abstract_scenario = AbstractScenario(self.input_config)

            # Update SUMO Network before generating OpenDrive network
            if demo_data := self.input_config["demo_data"]:

                # demo mode is enabled, use the updated SUMO network from demo data
                incl_sumo_net = pf.path2linux(Path(self.input_config["input_dir"]) / f"updated_net/{demo_data}.net.xml")

            if incl_sumo_net:
                # check if the file exists and end with .net.xml
                if incl_sumo_net.endswith(".net.xml") and os.path.exists(incl_sumo_net):
                    self.input_config["incl_sumo_net"] = incl_sumo_net

                    # Copy user updated net file to the OpenDrive folder
                    incl_sumo_net = pf.path2linux(Path(incl_sumo_net))  # ensure it's absolute path
                    if incl_sumo_net != path_sumo_net:
                        shutil.copy(incl_sumo_net, path_sumo_net)

                    # update opendrive network
                    self.abstract_scenario.Network.OpenDriveNetwork.OpenDrive_network = [
                        path_sumo_net, ""]

                    # mapping SUMO to OpenDrive IDs
                    parse_SUMO_ID(path_sumo_net)
                    parse_SUMO_to_OpenDrive(path_sumo_net)

                    console.print(f"  [dim cyan]:INFO: SUMO network is copied to {path_sumo_net}.\n"
                                  f"  [dim cyan]:Using updated SUMO network provide by user: {incl_sumo_net} "
                                  "to generate OpenDrive network.\n", soft_wrap=True, no_wrap=False)

                    # create opendrive net from updated sumo net, and rewrite sumo net based on OpenDrive net
                    # self.abstract_scenario.create_OpenDrive_network()
                    # rprint("  [dim cyan]:INFO: OpenDrive network is generated.", end="")
                else:
                    console.print("  [magenta]:NOTE: incl_sumo_net is not exist or not with .net.xml extension.\n"
                                  "  :Please provide a valid SUMO file with .net.xml extension or leave it empty.",
                                  soft_wrap=True, no_wrap=False)
            else:
                #  let user know they can use their own SUMO network by using incl_sumo_net
                rprint("  [dim cyan]:INFO: You can use your own SUMO network by providing the path "
                       "to the incl_sumo_net parameter. The path should be a .net.xml file. \n",
                       end="")

            # generate SUMO and OpenDrive network if not exists
            if not os.path.exists(path_sumo_net):
                # Create original SUMO network from vertices from config file
                self.abstract_scenario.create_SUMO_network()

                # crate OpenDrive network from SUMO network, and then rewrite sumo network based on OpenDrive network
                self.abstract_scenario.create_OpenDrive_network()

                rprint("  :INFO: OpenDrive network is generated.\n", end="")

            # create matchup table for user
            path_matchup = pf.path2linux(Path(self.input_config.get("input_dir")) / "MatchupTable.xlsx")

            # check if sumo net file exists
            if not os.path.exists(path_sumo_net):
                raise Exception(f"  :Error: SUMO net file does not exist: {path_sumo_net},"
                                "please check input configuration file and re-run the script."
                                "For details please refer to the documentation: ")
            df_matchup_table = format_junction_bearing(path_sumo_net)
            generate_matchup_table(df_matchup_table, path_matchup)
            console.print(f"  [dim cyan]:NOTE: Matchup table is generated and saved to {path_matchup}.[/dim cyan]\n"
                          "  :NOTICE: [bold red]Please update the Matchup table from input folder"
                          " and then run generate_abstract_scenario()."
                          " For details please refer to official documentation: \n", soft_wrap=True, no_wrap=False)

            if self._input_confirm:
                console.rule("[bold green]Program stopped. Please prepare the Control and Traffic data and "
                             "fill in the Matchup Table before proceeding.\n"
                             "[bold red]Please run generate_abstract_scenario() "
                             "after preparing the input data.\n")
                time.sleep(2)  # wait for 2 seconds before exiting
                # sys.exit(0)
                # Stop the program to let user update the Matchup table

                usr_input = False
                while not usr_input:
                    usr_input = console.input(":warning: [bold magenta]Please update the generated Matchup table from "
                                              "input folder before pressing Enter or type 'y' / 'yes' to continue")
                    if usr_input in {"", "y", "Y", "yes", "Yes"}:
                        console.print("  [dim cyan]:INFO: User confirmed to continue (Matchup Table Updated).")
                        usr_input = True

        return (f"  :Info: Please prepare the Control and Traffic data and save them to the input directory. "
                f"Please also fill in the Matchup Table at {path_matchup}.\n")

    def generate_abstract_scenario(self):
        """Generate the abstract scenario: create OpenDrive files
        """

        # check abstract scenario inputs
        check_abstract_inputs(self.input_config.get("input_dir"))

        # 1. Generate the abstract scenario based on the input data
        # self.abstract_scenario = AbstractScenario(self.input_config)
        if not hasattr(self, 'abstract_scenario'):
            raise Exception("  :Error: Abstract Scenario is not generated yet. "
                            "Please run generate_inputs() first.")

        # update traffic and signal
        path_matchup_updated = Path(self.input_config.get("input_dir")) / "MatchupTable_updated.xlsx"
        path_matchup = pf.path2linux(Path(self.input_config.get("input_dir")) / "MatchupTable.xlsx")
        if path_matchup_updated.exists():
            # update the matchup table with the updated one
            shutil.copy(path_matchup_updated, path_matchup)
        else:
            path_matchup = pf.path2linux(Path(self.input_config.get("input_dir")) / "MatchupTable.xlsx")
        print(f"  :INFO: Using Matchup Table: {path_matchup}")
        control_dir = pf.path2linux(Path(self.input_config.get("input_dir")) / "Control")
        traffic_dir = pf.path2linux(Path(self.input_config.get("input_dir")) / "Traffic")

        # Auto-fill matchup table, save to matchup table : MatchupTable_UserInput
        _ = update_matchup_table(path_matchup_table=path_matchup,
                                 control_dir=control_dir,
                                 traffic_dir=traffic_dir)

        if self._input_confirm:
            # Tell user to manually check correctness of the Matchup Table
            console.input(":warning: [bold magenta]In the Matchup Table, please check if the turn movement in the "
                          "demand and control data match with bearings in the network data. Enter any key to continue...")

        df_volume, df_vol_lookup = generate_turn_demand(path_matchup_table=path_matchup,
                                                        control_dir=control_dir,
                                                        traffic_dir=traffic_dir,)

        self.abstract_scenario.Traffic.VolumeLookupTable = df_vol_lookup

        self.abstract_scenario.update_AbstractScenario_from_input(df_volume=df_volume)

        # UPDATE: delete original TLLogic in the net.xml file and assign new TLS logic from the matchup table
        net_name = self.input_config["Network"]["NetworkName"]
        path_sumo_net = pf.path2linux(Path(self.input_config.get("output_dir")) / f"OpenDrive/{net_name}.net.xml")
        parse_SUMO_TLS_ID(path_matchup_table=path_matchup,
                          path_net_file=path_sumo_net)

        console.print("\n[bold green]Abstract Scenario successfully generated.")

    def generate_concrete_scenario(self):
        """Generate the concrete scenario: generate unified scenario from abstract scenario
        """

        # 1. Generate the concrete scenario based on abstract scenario
        # 2. Save the concrete scenario to the output directory

        if not hasattr(self, 'abstract_scenario'):
            console.print("  [magenta]:Warning: Abstract Scenario is not generated yet. "
                          "Please run generate_abstract_scenario() first.")
            return

        self.concrete_scenario = ConcreteScenario()
        self.concrete_scenario.get_unified_scenario(self.abstract_scenario)
        console.print("\n[bold green]Concrete Scenario successfully generated.")

    def prepare_simulation(self,
                           start_time: float = 3600 * 8,
                           end_time: float = 3600 * 10,
                           seed: list | int = 812,
                           step_length: float = 0.1) -> bool:
        """Simulate the concrete scenario: generate simulation files for the selected simulator

        Args:
            start_time (float): The start time of the simulation. Default is 3600 * 8.
            end_time (float): The end time of the simulation. Default is 3600 * 10.
            seed (list or int): The seed for the simulation. Default is [101].
            step_length (float): The simulation step size. Default is 0.1.

        Examples:
            import realtwin package
            >>> import realtwin as rt

            load the input configuration file
            >>> twin = rt.REALTWIN(input_config_file="config.yaml", verbose=True)

            check simulator is installed or not, default to SUMO
            >>> twin.env_setup(sel_sim=["SUMO"])

            generate abstract scenario and concrete scenario
            >>> twin.generate_abstract_scenario()
            >>> twin.generate_concrete_scenario()

            prepare simulation with start time, end time, seed, and step size
            >>> twin.prepare_simulation(start_time=3600 * 8, end_time=3600 * 10, seed=[101], step_length=0.1)

        Returns:
            bool: True if the simulation is prepared successfully, False otherwise.
        """

        # 1. prepare Simulate docs from the concrete scenario
        # 2. Save results to the output directory

        sim_prep = {
            "sumo": SimPrep().create_sumo_sim,
            "vissim": SimPrep().create_vissim_sim,
            "aimsun": SimPrep().create_aimsun_sim,
        }

        # TODO according sel_sim to run different simulators
        self.sim = SimPrep()
        for simulator in self.sel_sim:
            sim_prep.get(simulator)(self.concrete_scenario,
                                    start_time=start_time,
                                    end_time=end_time,
                                    seed=seed,
                                    step_length=step_length)
            console.print(f"\n[bold green]{simulator.upper()} simulation successfully Prepared.")
        return True

    def calibrate(self, *, sel_algo: dict = None,
                  sel_behavior_routes: dict = None,
                  update_turn_flow_algo: dict = None,
                  update_behavior_algo: dict = None) -> bool:
        # sourcery skip: extract-duplicate-method, remove-empty-nested-block, remove-redundant-if
        """Calibrate the turn and inflow, and behavioral parameters using the selected algorithms.

        Args:
            sel_algo (dict): The dictionary of algorithms to be used for calibration.
                Default is None, will use genetic algorithm. e.g. {"turn_inflow": "ga", "behavior": "ga"}.
            sel_behavior_routes (dict): The dictionary of behavior route parameters to be used for calibration.
                Default is None. time (in seconds) is ground truth travel time.
                e.g. sel_behavior_routes = {"route_1": {"time": 20, "edge_list": ["edge_id_1", "edge_d_2", ...]},
                                           "route_2" {"time": 40, "edge_list":["edge_id_1", "edge_d_2", ...]}
                                           ...}.
            update_turn_flow_algo (dict): The dictionary of algorithms to be used for updating turn flow.
                Default is None, will use genetic algorithm.
                Please refer to input configuration file for keys for each algorithm.
                e.g. update_turn_flow_algo = {"ga_config": {}, "sa_config":{}, "ts_config":{}}.
            update_behavior_algo (dict): The dictionary of algorithms to be used for updating behavior.
                Default is None, will use genetic algorithm.
                Please refer to input configuration file for keys for each algorithm.
                e.g. update_behavior_algo = {"ga_config":{}, "sa_config":{}, "ts_config": {}}.
        """
        # TDD
        print()
        if sel_algo is None:  # default to genetic algorithm
            sel_algo = {"turn_inflow": "ga", "behavior": "ga"}
            console.print(f"  [dim cyan]:sel_algo not specified, use default value: {sel_algo}")

        if not isinstance(sel_algo, dict):
            sel_algo = {"turn_inflow": "ga", "behavior": "ga"}
            console.print("  [bold red]:Error: parameter sel_algo must be a dict with"
                          " keys of 'turn_inflow' and 'behavior', using"
                          f" default values: {sel_algo}")

        # check if the selected algorithm is supported within the package
        # convert the algorithm to lower case
        sel_algo = {key: value.lower() for key, value in sel_algo.items()}
        if (algo := sel_algo["turn_inflow"]) not in ["ga", "sa", "ts"]:
            console.print(f"  [dim cyan]:Selected algorithms are {sel_algo}")
            console.print(f"  [dim cyan]:{algo} for turn and inflow calibration is not supported. "
                          "Must be one of ['ga', 'sa', 'ts']")
            return False

        if (algo := sel_algo["behavior"]) not in ["ga", "sa", "ts"]:
            console.print(f"  [dim cyan]:Selected algorithms are {sel_algo}")
            console.print(f"  [div cyan]:{algo} for behavior calibration is not supported. "
                          "Must be one of ['ga', 'sa', 'ts']")
            return False

        # parse user additional parameters for calibration
        user_kwargs = {}
        if sel_behavior_routes:
            # use user defined behavior route, if not provided, automatically select two routes from the network
            user_kwargs["sel_behavior_routes"] = sel_behavior_routes
        if self.input_config["demo_data"] and sel_behavior_routes_demo.get(self.input_config["demo_data"]):
            # use predefined behavior routes for demo data
            user_kwargs["sel_behavior_routes"] = sel_behavior_routes_demo.get(self.input_config["demo_data"])
        if update_turn_flow_algo:
            user_kwargs["update_turn_flow_algo"] = update_turn_flow_algo
        if update_behavior_algo:
            user_kwargs["update_behavior_algo"] = update_behavior_algo

        # run calibration based on the selected algorithm
        if "sumo" in self.sel_sim:
            cali_sumo(sel_algo=sel_algo, input_config=self.input_config, verbose=self.verbose, **user_kwargs)

        if "vissim" in self.sel_sim:
            pass

        if "aimsun" in self.sel_sim:
            pass

        console.print("[bold green]Calibration successfully completed.\n")

        return True

    def post_process(self):
        """Post-process the simulation results.
        """

        # 1. Post-process the simulation results
        # 2. Save the post-processed results to the output directory

    def visualize(self):
        """Visualize the simulation results.
        """

        # 1. Visualize the simulation results
        # 2. Save the visualization results to the output directory



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host a unique AbstractScenario definition
'''
import os
import warnings
import io
import copy
from pathlib import Path

# import four elements of AbstractScenario
from ._traffic import Traffic
from ._network import Network
from ._control import Control
from ._application import Application

import pandas as pd
import pyufunc as pf


def time_to_seconds(time_str) -> int:
    """convert time string to seconds

    Args:
        time_str (str): the time string in format 'HH:MM'
    """
    # hour, minute = map(int, time_str.split(':'))
    hour, minute = [int(x) for x in time_str.split(':')]
    return (hour * 3600) + (minute * 60)


def load_traffic_volume(demand_data: str | pd.DataFrame) -> pd.DataFrame:
    """load traffic volume data from file

    Args:
        demand_data (str): the demand file path

    Returns:
        pd.DataFrame: the converted demand data in DataFrame
    """
    # TDD check whether the file exists
    if not isinstance(demand_data, (str, pd.DataFrame)):
        warnings.warn(f"\n  :demand_data is not a string or dataframe: {demand_data}"
                      "\n  :No traffic volume data loaded from input file")
        return None

    if isinstance(demand_data, str):
        if not os.path.isfile(demand_data):
            warnings.warn(f"  :File not found: {demand_data}"
                          "\n  :No traffic volume data loaded from input file")
            return None

        # read the csv file and fill the nan values with 0
        traffic_volume = pd.read_csv(demand_data)
        traffic_volume.fillna(0, inplace=True)

        # Create a copy of the DataFrame
        df_volume = traffic_volume.copy()

    elif isinstance(demand_data, pd.DataFrame):
        # If demand_data is already a DataFrame, use it directly
        df_volume = demand_data.copy()

    # Apply the conversion function to the 'Time' column and create a new 'Seconds' column
    df_volume['IntervalStart'] = df_volume['Time'].apply(time_to_seconds)
    df_volume['IntervalEnd'] = df_volume['IntervalStart'] + 15 * 60
    df_volume = df_volume.drop('Time', axis=1)

    # Reshape the DataFrame to the long format
    df_volume = df_volume.melt(id_vars=['IntersectionName', 'IntervalStart', 'IntervalEnd'],
                               var_name='Turn',
                               value_name='Count')

    # clean Count column: change "" to 0 and convert to int
    df_volume['Count'] = df_volume['Count'].replace("", 0).astype(int)

    # Sort the DataFrame by IntersectionName and Turn columns
    df_volume.sort_values(['IntersectionName', 'IntervalStart', 'IntervalEnd'], inplace=True)

    # Reset the index
    df_volume.reset_index(drop=True, inplace=True)

    return df_volume


def load_traffic_turning_ratio(df_volume: pd.DataFrame) -> pd.DataFrame:
    """load traffic turning ratio data from file

    Args:
        df_volume (pd.DataFrame): the volume data in DataFrame

    Returns:
        pd.DataFrame: the converted turning ratio data in DataFrame
    """

    TurnDfTemp = copy.deepcopy(df_volume)

    # NBT to N and T
    TurnDfTemp['Bound'] = TurnDfTemp['Turn'].str[0]
    TurnDfTemp['Direction'] = TurnDfTemp['Turn'].str[-1]
    TurnDfTemp["IntervalStart"] = TurnDfTemp["IntervalStart"].astype(str)
    TurnDfTemp["IntervalEnd"] = TurnDfTemp["IntervalEnd"].astype(str)
    FlowTemp = TurnDfTemp.groupby(['IntervalStart', 'IntervalEnd', 'IntersectionName', 'Bound'],
                                  as_index=False)['Count'].sum()
    df_turning_ratio = pd.merge(TurnDfTemp, FlowTemp,
                                on=['IntervalStart', 'IntervalEnd',
                                    'IntersectionName', 'Bound'],
                                how='left')
    df_turning_ratio['TurnRatio'] = df_turning_ratio['Count_x'] / \
        df_turning_ratio['Count_y']
    df_turning_ratio = df_turning_ratio.drop(['Count_x', 'Count_y'], axis=1)
    df_turning_ratio.reset_index(drop=True, inplace=True)

    return df_turning_ratio


def load_control_signal(path_signal: str) -> dict:
    """load control signal data from file

    Args:
        path_signal (str): the signal file path

    Returns:
        dict: the converted signal data in dictionary
    """

    # TDD check whether the file exists
    if not isinstance(path_signal, str):
        warnings.warn(f"\n  :File path is not a string: {path_signal}"
                      "\n  :No signal data loaded from input file")
        return None

    if not os.path.isfile(path_signal):
        warnings.warn(f"  :File not found: {path_signal}")
        return None

    # read the signal file
    with open(path_signal, 'r', encoding="utf-8") as file:
        signal = file.readlines()

    SignalDict = {}
    current_table = None
    current_table_data = []
    # Iterate over the lines
    remove_flag = 0
    for line in signal:
        line = line.strip()

        # Check if it's a line to be deleted
        if remove_flag == 1:
            remove_flag = 0
            continue

        # Check if it's a table name
        if line.startswith("["):
            remove_flag = 1
            if current_table is None:
                current_table = line[1:-1]  # Remove the square brackets
            else:
                # Store the previous table data in the dictionary
                SignalDict[current_table] = pd.read_csv(io.StringIO('\n'.join(current_table_data)))

                # Start a new table
                current_table = line[1:-1]  # Remove the square brackets
                current_table_data = []
        else:
            current_table_data.append(line)

    # Store the last table in the dictionary
    SignalDict[current_table] = pd.read_csv(io.StringIO('\n'.join(current_table_data)))

    return SignalDict


class AbstractScenario:
    """Initialize an Abstract Scenario"""

    def __init__(self, input_config: dict = None):

        self.input_config = input_config

        self.Traffic = Traffic()
        self.Network = Network(output_dir=self.input_config.get('output_dir'))

        # update Network
        if network_dict := self.input_config.get('Network'):
            self.Network.NetworkName = network_dict.get('NetworkName', "network")
            self.Network.NetworkVertices = network_dict.get('NetworkVertices', "")
            self.Network.ElevationMap = network_dict.get('ElevationMap', "No elevation map provided!")

            # update the OpenDriveNetwork output directory
            self.Network._output_dir = self.input_config.get('output_dir')
            self.Network.OpenDriveNetwork._output_dir = self.Network._output_dir

            # update and crate OpenDriveNetwork
            self.Network.OpenDriveNetwork._net_name = self.Network.NetworkName
            self.Network.OpenDriveNetwork._net_vertices = self.Network.NetworkVertices
            self.Network.OpenDriveNetwork._ele_map = self.Network.ElevationMap

        self.Control = Control()
        self.Application = Application()

    def create_SUMO_network(self):
        """ Create SUMO Network From Vertices"""

        # TDD check whether the input_config is not None
        if not self.input_config:
            warnings.warn("  :input_config is None, no data to update")
            return

        # update Network
        # network_dict = self.input_config.get('Network', None)
        if network_dict := self.input_config.get('Network'):
            self.Network.NetworkName = network_dict.get('NetworkName', "network")
            self.Network.NetworkVertices = network_dict.get('NetworkVertices', "")
            self.Network.ElevationMap = network_dict.get('ElevationMap', "No elevation map provided!")

            # update the OpenDriveNetwork output directory
            self.Network._output_dir = self.input_config.get('output_dir')
            self.Network.OpenDriveNetwork._output_dir = self.Network._output_dir

            # update and crate OpenDriveNetwork
            self.Network.OpenDriveNetwork._net_name = self.Network.NetworkName
            self.Network.OpenDriveNetwork._net_vertices = self.Network.NetworkVertices
            self.Network.OpenDriveNetwork._ele_map = self.Network.ElevationMap
            self.Network.OpenDriveNetwork.create_SUMO_network()

    def create_OpenDrive_network(self):
        """create OpenDriveNetwork object"""
        # TDD check whether the input_config is not None
        if not self.input_config:
            warnings.warn("  :input_config is None, no data to update")
            return

        self.Network.OpenDriveNetwork.create_OpenDrive_network()

        # re-write sumo network based on the OpenDriveNetwork
        net_name = self.Network.OpenDriveNetwork._net_name
        path_open_drive = pf.path2linux(Path(self.Network._output_dir) / f"OpenDrive/{net_name}.xodr")
        path_sumo_net = pf.path2linux(Path(self.Network._output_dir) / f"OpenDrive/{net_name}.net.xml")
        os.system(f'cmd/c "netconvert --opendrive {path_open_drive}'
                  f' -o {path_sumo_net} --no-internal-links"')

    def update_AbstractScenario_from_input(self, df_volume: pd.DataFrame = None, signal_dict: dict = None):
        """ update values from config dict to specific data object"""

        # TDD check whether the input_config is not None
        if not self.input_config:
            warnings.warn("  :input_config is None, no data to update")
            return

        # # update Network
        # # network_dict = self.input_config.get('Network', None)
        # if network_dict := self.input_config.get('Network'):
        #     self.Network.NetworkName = network_dict.get('NetworkName', "network")
        #     self.Network.NetworkVertices = network_dict.get('NetworkVertices', "")
        #     self.Network.ElevationMap = network_dict.get('ElevationMap', "No elevation map provided!")
        #     # update the OpenDriveNetwork output directory
        #     self.Network._output_dir = self.input_config.get('output_dir', "RT_Network")
        #     self.Network.OpenDriveNetwork._output_dir = self.Network._output_dir
        #     # update and crate OpenDriveNetwork
        #     self.Network.OpenDriveNetwork._net_name = self.Network.NetworkName
        #     self.Network.OpenDriveNetwork._net_vertices = self.Network.NetworkVertices
        #     self.Network.OpenDriveNetwork._ele_map = self.Network.ElevationMap
        #     self.Network.OpenDriveNetwork.setValue()

        # update Traffic
        if df_volume is not None:
            self.Traffic.Volume = load_traffic_volume(df_volume)
            self.Traffic.TurningRatio = load_traffic_turning_ratio(self.Traffic.Volume)
        elif traffic_dict := self.input_config.get('Traffic'):
            path_volume = traffic_dict.get('Volume', None)
            path_volume_abs = pf.path2linux(os.path.join(self.input_config.get("input_dir"), path_volume))
            self.Traffic.Volume = load_traffic_volume(path_volume_abs)
            self.Traffic.TurningRatio = load_traffic_turning_ratio(self.Traffic.Volume)

        # update Control
        if signal_dict is not None:
            self.Control.Signal = signal_dict
        elif control_data := self.input_config.get('Control'):
            if isinstance(control_data, dict):
                path_signal = control_data.get('Signal', None)
                path_signal_abs = pf.path2linux(os.path.join(self.input_config.get("input_dir"), path_signal))
                self.Control.Signal = load_control_signal(path_signal_abs)
            elif isinstance(control_data, str):
                path_signal = control_data
                path_signal_abs = pf.path2linux(Path(self.input_config.get("input_dir")) / "Control" / path_signal)
                if os.path.isfile(path_signal_abs):
                    self.Control.Signal = load_control_signal(path_signal_abs)
                else:
                    raise FileNotFoundError(
                        f"  :File not found: {path_signal_abs}. No signal data loaded from input file")
            else:
                raise ValueError("  :Invalid control data in configuration file. ")

    def fillAbstractScenario(self):
        """Fill the AbstractScenario with the data from the input_config"""
        pass



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Utility functions for SUMO calibration and behavior modeling. """

import os
import sys
import xml.etree.ElementTree as ET
import subprocess
import warnings
import pandas as pd
import numpy as np
import folium
import sumolib
from itertools import combinations
import requests

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
    sys.path = list(set(sys.path))  # remove duplicates
else:
    warnings.warn("Environment variable 'SUMO_HOME' is not set. "
                  "please declare environment variable 'SUMO_HOME'")
    # sys.exit("please declare environment variable 'SUMO_HOME'")
import traci


def update_turn_flow_from_solution(path_turn: str,
                                   path_inflow: str,
                                   initial_solution: np.array,
                                   cali_interval: int,
                                   demand_interval: int) -> tuple:
    """assign the new turn ratios and inflow counts to the given dataframes

    Args:
        path_turn (str): the turn file path from turn.xlsx
        path_inflow (str): the inflow file path from inflow.xlsx
        initial_solution (np.array): the initial solution from the genetic algorithm
        cali_interval (int): the calibration interval
        demand_interval (int): the demand interval

    Returns:
        tuple: the updated turn and inflow dataframes
    """

    # Much improved in terms of speed and readability. 1 min for 2 - 5 size and generations

    # create the copy of turn and inflow dataframes for internal operations
    TurnDf = pd.read_excel(path_turn)
    InflowDf = pd.read_excel(path_inflow)

    # --- Update TurnRatios ---
    # Instead of many .loc calls, set a MultiIndex on the two key columns.
    TurnDf.set_index(['OpenDriveFromID', 'OpenDriveToID'], inplace=True)
    TurnDf.sort_index(inplace=True)

    # Create a mapping from (from, to) to the new TurnRatio.
    # Notice that for each pair one of the assignments is the value,
    # and the other is 1 minus that value.
    turn_mapping = {
        # Between Amin Dr. and  I-75 SB Off Ramp
        (290, 298): initial_solution[0],
        (290, 299): 1 - initial_solution[0],
        (331, 297): initial_solution[1],
        (331, 298): 1 - initial_solution[1],
        (293, 299): initial_solution[2],
        (293, 297): 1 - initial_solution[2],

        # Between Napier Rd. and Lifestyle Way1
        (315, 321): initial_solution[3],
        (315, 323): 1 - initial_solution[3],
        (320, 3221): initial_solution[4],
        (320, 321): 1 - initial_solution[4],
        (302, 323): initial_solution[5],
        (302, 3221): 1 - initial_solution[5],

        # Between Napier Rd. and Lifestyle Way2
        (281, 315): initial_solution[6],
        (281, 314): 1 - initial_solution[6],
        (316, 313): initial_solution[7],
        (316, 315): 1 - initial_solution[7],
        (322, 314): initial_solution[8],
        (322, 313): 1 - initial_solution[8],

        # Between  Lifestyle Way and Gunbarrel Road
        (330, 327): initial_solution[9],
        (330, 328): 1 - initial_solution[9],
        (307, 329): initial_solution[10],
        (307, 327): 1 - initial_solution[10],
        (284, 328): initial_solution[11],
        (284, 329): 1 - initial_solution[11],
    }

    # Loop over the (small) mapping dictionary.
    # Because the DataFrame index is a MultiIndex, each lookup is fast.
    for key, value in turn_mapping.items():
        if key in TurnDf.index:
            TurnDf.loc[key, 'TurnRatio'] = value

    # Reset the index so that the returned DataFrame has the original format.
    TurnDf.reset_index(inplace=True)

    # --- Update Inflow Counts ---
    # First ensure that Count is a float.
    InflowDf['Count'] = InflowDf['Count'].astype(float)

    # For inflows, the key is just OpenDriveFromID. Set it as the index.
    InflowDf.set_index('OpenDriveFromID', inplace=True)

    #  considering the calibration interval and demand interval
    inflow_mapping = {
        331: initial_solution[12] / cali_interval * demand_interval,
        320: initial_solution[13] / cali_interval * demand_interval,
        316: initial_solution[14] / cali_interval * demand_interval,
        330: initial_solution[15] / cali_interval * demand_interval,
    }

    # Loop over the inflow mapping and assign new values.
    for key, value in inflow_mapping.items():
        if key in InflowDf.index:
            InflowDf.loc[key, 'Count'] = value

    InflowDf.reset_index(inplace=True)
    return (TurnDf, InflowDf)


def run_SUMO_create_EdgeData(sim_name: str, sim_end_time: float) -> bool:
    """run SUMO simulation using traci module

    Args:
        sim_name (str): the name of the simulation, it should be the .sumocfg file
        sim_end_time (float): the end time of the simulation

    Returns:
        bool: True if the simulation is successful
    """

    traci.start(["sumo", "-c", sim_name])
    while traci.simulation.getTime() < sim_end_time:
        traci.simulationStep()
    traci.close()
    return True


def get_travel_time_from_EdgeData_xml(path_EdgeData: str, edge_ids: list) -> float:
    """
    Calculate total travel time along a route composed of multiple edges.

    Parameters:
    edge_output_file (str): Path to the edge output file generated by SUMO.
    edge_ids (list): List of edge IDs that make up the route.

    Returns:
    float: Total travel time along the route.
    """
    total_travel_time = 0.0
    tree = ET.parse(path_EdgeData)
    root = tree.getroot()
    # print (root.attrib)
    for edge_id in edge_ids:
        p = root.findall('interval')
        for parent in p:
            for child in parent:
                if child.get('id') == edge_id:
                    travel_time = child.get('traveltime')
                    if travel_time is not None:
                        total_travel_time += float(travel_time)
    return total_travel_time


def update_flow_xml_from_solution(path_flow: str, solution: list | np.ndarray) -> bool:
    """Update the flow XML file with new car-following parameters."""

    min_gap, accel, decel, sigma, tau, emergencyDecel = solution

    # Load the XML file
    tree = ET.parse(path_flow)
    root = tree.getroot()

    # Find the tag
    parent = root.find('vType')

    if parent is not None:
        # print(parent.tag, parent.attrib)  # Prints child tag name and text
        parent.set('minGap', str(min_gap))
        parent.set('accel', str(accel))  # Add a new attribute
        parent.set('decel', str(decel))
        parent.set('sigma', str(sigma))
        parent.set('tau', str(tau))
        parent.set('emergencyDecel', str(emergencyDecel))
    else:
        print("Parent tag not found")
    tree.write(path_flow)
    return True


def run_jtrrouter_to_create_rou_xml(network_name: str, path_net: str, path_flow: str, path_turn: str, path_rou: str, verbose: bool = False) -> None:
    """Runs jtrrouter to generate a route file from flow and network files in SUMO.

    Args:
        network_name (str): The name of the network.
        path_net (str): The path to the network file.
        path_flow (str): The path to the flow file.
        path_turn (str): The path to the turn file.
        path_rou (str): The path to the output route file.
        verbose (bool): If True, print additional information. Defaults to False.
    """

    # Define the jtrrouter command with all necessary arguments
    # cmd = [
    #     "jtrrouter",
    #     "-n", path_net,
    #     "-r", path_flow,
    #     "-t", path_turn,
    #     "-o", path_rou,
    #     "--accept-all-destinations",
    #     "--remove-loops True",
    #     # "--seed","101",
    #     "--ignore-errors",  # Continue on errors; remove if not desired
    # ]
    # cmd = f'jtrrouter -r "{path_flow}" -t "{path_turn}" -n "{path_net}" --accept-all-destinations --remove-loops True --randomize-flows -o "{path_rou}"'
    cmd = (
        f'jtrrouter -r "{path_flow}" '
        f'-t "{path_turn}" '
        f'-n "{path_net}" '
        f'--accept-all-destinations '
        f'--remove-loops True '
        f'--randomize-flows '
        f'-o "{path_rou}"'
    )

    # Execute the command
    try:
        # subprocess.run(cmd, capture_output=True, text=True)
        process = subprocess.Popen(cmd, shell=True)
        process.wait()
        if verbose:
            print(f"  :Route file generated successfully: {path_rou}")
    except subprocess.CalledProcessError as e:
        print(f"  :An error occurred while running jtrrouter: {e}")


def result_analysis_on_EdgeData(path_summary: str,
                                path_EdgeData: str,
                                calibration_target: dict,
                                sim_start_time: float,
                                sim_end_time: float) -> tuple:
    """Analyze the result of the simulation and return the flag, mean GEH, and GEH percent

    Args:
        path_summary (str or pd.DataFrame): the summary dataframe from summary.xlsx in input dir
        path_EdgeData (str): the path to the EdgeData.xml file in the input dir
        calibration_target (dict): the calibration target from the scenario config, it should contain GEH and GEHPercent
        sim_start_time (float): the start time of the simulation
        sim_end_time (float): the end time of the simulation

    Returns:
        tuple: (flag, mean GEH, geh percent)
    """
    # Load and parse the new XML file
    # mapping of sumo id with GridSmart Intersection from user input

    # 1. Filter and group the summary data
    df_summary = pd.read_excel(path_summary)
    df_filtered = df_summary.loc[df_summary["realcount"].notna()].copy()
    approach_summary = df_filtered.groupby(['IntersectionName',
                                            'entrance_sumo',
                                            'Bound'], as_index=False)['realcount'].sum()

    # 2. Parse the XML file using a list comprehension with immediate type conversion
    tree = ET.parse(path_EdgeData)
    root = tree.getroot()
    edge_data = [
        {
            'id': int(edge.get('id')) if edge.get('id') else None,
            'travel_time': float(edge.get('traveltime')) if edge.get('traveltime') else None,
            'arrived': int(edge.get('arrived')) if edge.get('arrived') else None,
            'departed': int(edge.get('departed')) if edge.get('departed') else None,
            'left': int(edge.get('left')) if edge.get('left') else None,
            'density': float(edge.get('density')) if edge.get('density') else None,
            'speed': float(edge.get('speed')) if edge.get('speed') else None
        }
        for interval in root.findall('.//interval')
        for edge in interval.findall('edge')
    ]
    edge_df = pd.DataFrame(edge_data)

    # 3. Ensure matching key types and merge data
    approach_summary['entrance_sumo'] = approach_summary['entrance_sumo'].astype(int)
    edge_df['id'] = edge_df['id'].astype(int)
    merged = approach_summary.merge(edge_df,
                                    left_on='entrance_sumo',
                                    right_on='id',
                                    how='inner')
    merged.rename(columns={'left': 'count'}, inplace=True)
    merged.drop(columns=['id'], inplace=True)

    # 4. Calculate flows (vehicles per hour)
    duration = sim_end_time - sim_start_time
    merged['flow'] = merged['count'] / duration * 3600
    merged['realflow'] = merged['realcount'] / duration * 3600

    # 5. Compute GEH and summary statistics
    merged['GEH'] = np.sqrt(2
                            * ((merged['count'] - merged['realcount']) ** 2)
                            / (merged['count'] + merged['realcount']))
    mean_geh = merged['GEH'].mean()
    geh_percent = (merged['GEH'] < calibration_target['GEH']).mean()

    flag = 1
    if geh_percent < calibration_target['GEHPercent']:
        flag = 0

    # 6. Compute absolute differences and relative differences once
    diff_abs = (merged['realflow'] - merged['flow']).abs()
    relative_diff = (
        (merged['realflow'] - merged['flow']) / merged['realflow']).abs()

    # 7. Vectorized condition checks for different volume ranges

    # within 100 vph for volumes < 700
    cond_low = (merged['realflow'] < 700) & (diff_abs > 100)

    # within 15%  for volumes  700-2700
    cond_mid = (merged['realflow'].between(700, 2700)) & (relative_diff > 0.15)

    # within 400 vph for volumes > 2700
    cond_high = (merged['realflow'] > 2700) & (diff_abs > 400)

    # If any of the conditions are met, set flag to 0
    if cond_low.any() or cond_mid.any() or cond_high.any():
        flag = 0

    return (flag, mean_geh, geh_percent)


def compute_route_summary(rou_file: str, net_file: str) -> pd.DataFrame:
    """
    Parses the SUMO .rou.xml and .net.xml files, counts each unique route's frequency,
    computes its total length, and returns a DataFrame sorted by descending frequency.
    """
    # 1) load network
    net = sumolib.net.readNet(net_file)

    # 2) parse the .rou.xml
    tree = ET.parse(rou_file)
    root = tree.getroot()

    # 3) accumulate counts and lengths
    route_counts = {}
    route_lengths = {}
    for route_elem in root.findall(".//route"):
        edges = tuple(route_elem.attrib["edges"].split())
        # freq
        route_counts[edges] = route_counts.get(edges, 0) + 1
        # length
        if edges not in route_lengths:
            route_lengths[edges] = sum(
                net.getEdge(e).getLength()
                for e in edges
                if net.getEdge(e) is not None
            )

    # 4) build DataFrame
    df = pd.DataFrame([
        {
            "route": " ".join(edges),
            "frequency": freq,
            "length_meters": route_lengths[edges]
        }
        for edges, freq in route_counts.items()
    ])

    # 5) sort and return
    return df.sort_values("frequency", ascending=False).reset_index(drop=True)


def filter_mid_routes(df_sorted: pd.DataFrame) -> pd.DataFrame:
    """Filter 80-90th percentile by length"""

    length_85 = df_sorted["length_meters"].quantile(0.85)
    length_95 = df_sorted["length_meters"].quantile(0.95)
    return df_sorted[(df_sorted["length_meters"] >= length_85)
                     & (df_sorted["length_meters"] < length_95)].reset_index(drop=True)


def select_two_distinct(mid_routes: pd.DataFrame) -> list[list[str]]:
    """Select two most distinct via Jaccard """

    def jaccard_dist(a: str, b: str) -> float:
        sa, sb = set(a.split()), set(b.split())
        return 1 - len(sa & sb) / len(sa | sb)

    pairs = list(combinations(range(len(mid_routes)), 2))
    scores = [
        (i, j, jaccard_dist(mid_routes.loc[i, "route"], mid_routes.loc[j, "route"]))
        for i, j in pairs
    ]
    i, j, _ = max(scores, key=lambda x: x[2])
    return [
        mid_routes.loc[i, "route"].split(),
        mid_routes.loc[j, "route"].split()
    ]


def get_route_coords(net, edge_list: list[str]) -> list[tuple[float, float]]:
    """Extract WGS84 coords for a route
    net: a sumolib Net object
    edge_list: list of edge IDs (strings)
    """
    pts = []
    for eid in edge_list:
        edge = net.getEdge(eid)
        for x, y in edge.getShape():
            lon, lat = net.convertXY2LonLat(x, y)
            pts.append((lat, lon))
    return pts


def estimate_travel_time(route_coords: list[tuple[float, float]], api_key: str = "") -> int | None:
    """Estimate travel time with waypoint downsampling"""
    if len(route_coords) < 2:
        return None

    if not api_key:
        api_key = "AIzaSyBH5lSmawDWrJpPw9cd4h6wgN04lGNt3_A"

    def downsample(C: list, max_pts=23) -> list:
        if len(C) <= max_pts:
            return C
        step = max(1, len(C) // max_pts)
        return C[::step][:max_pts]

    origin = f"{route_coords[0][0]},{route_coords[0][1]}"
    destination = f"{route_coords[-1][0]},{route_coords[-1][1]}"
    mids = downsample(route_coords[1:-1], 23)
    waypts = "|".join(f"{lat},{lon}" for lat, lon in mids)

    resp = requests.get(
        "https://maps.googleapis.com/maps/api/directions/json",
        params={
            "origin": origin,
            "destination": destination,
            "waypoints": waypts,
            "departure_time": "now",
            "key": api_key
        },
        timeout=200
    )
    j = resp.json()
    if j.get("status") == "OK":
        return sum(leg["duration"]["value"] for leg in j["routes"][0]["legs"])
    print("API error:", j.get("status"), j.get("error_message"))
    return None


def plot_and_report(route_coords_list: list[list[tuple[float, float]]],
                    travel_times: list[int | None],
                    selected_routes: list[list[str]],
                    out_html: str = "routes_travel_time_map.html"):

    """ Plot routes, print edges & times, and add legend"""
    # Print edge lists and travel times
    for idx, edges in enumerate(selected_routes, start=1):
        print(f"Route {idx} edges: {edges}")
        t = travel_times[idx - 1]
        if t is not None:
            print(f"Route {idx} travel time: {t} sec ({t / 60:.1f} min)")
        else:
            print(f"Route {idx} travel time: not available")

    # Folium map
    all_pts = [pt for coords in route_coords_list for pt in coords]
    center = (
        sum(lat for lat, lon in all_pts) / len(all_pts),
        sum(lon for lat, lon in all_pts) / len(all_pts)
    )
    m = folium.Map(location=center, zoom_start=14)
    colors = ["red", "blue"]

    for idx, coords in enumerate(route_coords_list, start=1):
        tooltip = f"Route {idx}"
        if travel_times[idx - 1] is not None:
            tooltip += f": {travel_times[idx - 1]} sec"
        folium.PolyLine(coords, color=colors[idx - 1], weight=5, opacity=0.8,
                        tooltip=tooltip).add_to(m)

    # Legend
    legend_html = """
     <div style="
       position: fixed;
       bottom: 50px; left: 50px; width: 150px; height: 80px;
       background-color: white; border:2px solid grey; z-index:9999;
       font-size:14px; line-height:18px;
       ">
       &nbsp;<b>Route Legend</b><br>
       &nbsp;<span style="color:red;">&#8212;&#8212;</span>&nbsp;Route 1<br>
       &nbsp;<span style="color:blue;">&#8212;&#8212;</span>&nbsp;Route 2
     </div>
    """
    m.get_root().html.add_child(folium.Element(legend_html))

    m.save(out_html)
    print(f"Map saved → {out_html}")


def auto_select_two_routes(path_rou: str, path_net: str, api_key: str = "",
                           path_report: str = "routes_travel_time_map.html") -> tuple:
    """Automatically select two distinct routes from the given route file and network file.
    Using Google API to estimate travel time.

    Args:
        path_rou (str): sumo .rou.xml file
        path_net (str): sumo .net.xml file
        path_report (str): save the map to this path
        api_key (str, optional): Google Api key. Defaults to "".

    Returns:
        tuple: time and edge id list for the two routes
    """

    df_sorted = compute_route_summary(path_rou, path_net)
    mid_df = filter_mid_routes(df_sorted)
    routes = select_two_distinct(mid_df)

    net = sumolib.net.readNet(path_net)
    coords_list = [get_route_coords(net, route) for route in routes]
    time_list = [estimate_travel_time(coords, api_key) for coords in coords_list]
    plot_and_report(coords_list, time_list, routes, path_report)
    return (routes, time_list, coords_list)



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host demand element of Concrete scenario
'''
import os
import xml.etree.ElementTree as ET
import pandas as pd
import pyufunc as pf


class Demand:
    '''The demand class to host the demand element of Concrete scenario
    '''
    def __init__(self):
        self.Inflow = {}
        # Inflow: ['BeginInterval', 'EndInterval', 'FromEdge', 'Count', 'VehicleType']
        # TurningRatio: ['BeginInterval', 'EndInterval', 'FromEdge','ToEdge', 'Bound','TurnPercent','TurnDirection']

    def is_empty(self):
        """Check if the Demand object is empty."""
        pass

    def generate_traffic(self, AbsScn):
        '''Generate the traffic demand for the concrete scenario'''

        Count = AbsScn.Traffic.Volume
        NetworkPath = AbsScn.Network.OpenDriveNetwork.OpenDrive_network[0]
        # Read this for now

        if not AbsScn.Traffic.VolumeLookupTable.empty:
            IDRef = AbsScn.Traffic.VolumeLookupTable
        else:
            path_lookup = AbsScn.input_config["Traffic"].get("GridSmart_lookup")
            path_lookup_abs = pf.path2linux(os.path.join(AbsScn.input_config.get("input_dir"), path_lookup))
            IDRef = pd.read_csv(path_lookup_abs)
        IDRef = IDRef.dropna(subset=['OpenDriveFromID', 'OpenDriveToID'])
        # IDRef = IDRef.astype({'OpenDriveFromID': int, 'OpenDriveToID': int})
        IDRef = IDRef.astype(str)

        MergedDf1 = pd.merge(Count, IDRef, on=['IntersectionName', 'Turn'], how='left')
        Count['OpenDriveFromID'] = MergedDf1['OpenDriveFromID']
        Count['OpenDriveToID'] = MergedDf1['OpenDriveToID']
        Count = Count.groupby(['IntervalStart',
                               'IntervalEnd',
                               'IntersectionName',
                               'OpenDriveFromID'], as_index=False)['Count'].sum()

        Count = Count.dropna(subset=['OpenDriveFromID'])
        Count = Count.astype({'Count': int})
        # Omit exiting approach to avoid replicate problem for not
        # Count2 = Count.groupby(['IntervalStart', 'IntervalEnd', 'IntersectionName',
        # 'OpenDriveToID'], as_index=False)['Count'].sum()

        # Load the XML file
        tree = ET.parse(NetworkPath)
        root = tree.getroot()

        # Find all roads with the specified condition
        InflowRoad = []

        for road in root.findall('road'):
            link = road.find('link')
            if link is not None:
                predecessor = link.find('predecessor')
                if predecessor is not None and predecessor.get('elementType') == 'junction':
                    junction_id = predecessor.get('elementId')

                    # Find the junction element with the specified id
                    junction = root.find(f".//junction[@id='{junction_id}']")

                    # Check if the junction has only one connection
                    if junction is not None and len(junction.findall('connection')) == 1:
                        road_id = road.get('id')
                        InflowRoad.append(road_id)

        InflowCount = Count[Count['OpenDriveFromID'].isin(InflowRoad)]

        self.Inflow = InflowCount.copy()



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""Load configuration for autonomous vehicles in SUMO."""

import contextlib
import yaml
from pathlib import Path
try:
    from ._carfollowing_lanechanging_model import CFmodel, CF_DEFAULT_PARAMETERS
except ImportError:
    from realtwin.autonomous_veh._carfollowing_lanechanging_model import CFmodel, CF_DEFAULT_PARAMETERS


def load_av_configs(path_config: str | Path) -> dict:
    """Load the autonomous vehicle configuration from a YAML file.

    Args:
        path_config (str): Path to the configuration file.

    Returns:
        dict: Configuration parameters loaded from the file.
    """

    if isinstance(path_config, (str, Path)):
        path_config = str(path_config)
    else:
        raise TypeError("path_config must be a string or Path object.")

    # TDD: check if the file exists and is a valid YAML file
    if not Path(path_config).is_file():
        raise FileNotFoundError(f"Configuration file not found: {path_config}")

    if not (path_config.endswith('.yml') or path_config.endswith('.yaml')):
        raise ValueError("Configuration file must be a YAML file with .yml or .yaml extension.")

    # load the YAML file
    with open(path_config, 'r', encoding="utf-8") as yaml_data:
        config = yaml.safe_load(yaml_data)

    # check veh penetration
    pct_penetration = config.get('pct_penetration')
    if not isinstance(pct_penetration, list):
        raise TypeError("pct_penetration must be a list of percentages.")
    if sum(pct_penetration) != 100:
        raise ValueError("pct_penetration must sum to 100%.")
    for pct in pct_penetration:
        if (pct < 0) or (pct > 100):
            raise ValueError("pct_penetration values must be between 0 and 100.")

    # check veh types, from user defined veh_types
    veh_types = config.get('veh_types', [])
    if not veh_types:
        raise ValueError("Vehicle types must be specified in the configuration file.")

    CFmodel_USER = {
        veh_type: CF_DEFAULT_PARAMETERS.copy()
        for veh_type in veh_types
        if veh_type not in CFmodel
    }

    for veh_type in veh_types:
        if veh_type not in CFmodel:
            CFmodel[veh_type] = CF_DEFAULT_PARAMETERS.copy()

    # update the CFmodel parameters for each veh_type
    CF_model_names = list(CF_DEFAULT_PARAMETERS.keys())

    # for i in range(len(veh_types)):
    for i, veh_type in enumerate(veh_types):
        # copy the default CFmodel for each veh_type
        CFmodel_USER[veh_type] = CFmodel[veh_type].copy()

        # update default parameters with user defined parameters
        for model_name in CF_model_names:
            # get user defined parameters for each model
            model_key = f"{model_name}Parameters"
            user_model = config.get(model_key, {})

            # update each parameter for the veh_type
            for model_param in user_model:
                with contextlib.suppress(IndexError):
                    # get the parameter value from user input
                    param_values = user_model[model_param][i]
                    CFmodel_USER[veh_type][model_name][model_param] = param_values
            # delete the model_key from config to avoid confusion
            if model_key in config:
                del config[model_key]

    # Add the updated CFmodel to the config dictionary for future use
    config["CFmodel"] = CFmodel_USER

    return config



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

import os
import sys
import subprocess
import shutil
import pyufunc as pf


def venv_create(*, venv_name: str = "", venv_dir: str = "", pkg_name: str = "realtwin", verbose: bool = True) -> bool:
    """Create a virtual environment in the specified folder with the specified name.

    Args:
        venv_name (str): the name of the virtual environment
        venv_dir (str): the path to the folder where the virtual environment will be created
        pkg_name (str): the name of the package to be installed in the virtual environment
        verbose (bool): whether to print the progress

    Example:
        >>> import realtwin as rt
        >>> rt.venv_create(venv_name="my_venv", venv_dir="/path/to/dir")

    Raises:
        Exception: if env_name is not a string, or folder_path is not a string

    Returns:
        bool: True if the virtual environment is created successfully, False otherwise
    """
    # Default values for env_name and folder_path if not provided
    if not venv_name:
        venv_name = "venv_rt"

    if not venv_dir:
        venv_dir = os.getcwd()

    # TDD for venv_name
    if not isinstance(venv_name, str):
        raise Exception("env_name must be a string")
    if not isinstance(venv_dir, str):
        raise Exception("folder_path must be a string")

    # Create the virtual environment
    try:
        # Ensure the folder exists
        os.makedirs(venv_dir, exist_ok=True)

        # Full path to the virtual environment
        venv_path = pf.path2linux(os.path.join(venv_dir, venv_name))

        # Create the virtual environment
        subprocess.check_call([sys.executable, "-m", "venv", venv_path])
        # subprocess.run(["python", "-m", "venv", venv_path], check=True)

        if verbose:
            print(f"  :Virtual environment {venv_name} created at: {venv_path}")
    except subprocess.CalledProcessError as e:
        print(f"  :Failed to create virtual environment. Error: {e}")
        return False

    # Find the executable python in the virtual environment
    if pf.is_windows():
        python_executable = pf.path2linux(os.path.join(venv_path, 'Scripts', 'python.exe'))
        # activate_path = pf.path2linux(os.path.join(venv_path, "Scripts", "activate"))
        # subprocess.run([activate_path], shell=True, check=True)

    elif pf.is_linux() or pf.is_mac():
        python_executable = pf.path2linux(os.path.join(venv_path, 'bin', 'python'))
        # activate_path = os.path.join(venv_path, "bin", "activate")
        # subprocess.run(["source", activate_path], shell=True, check=True)

    else:
        raise Exception("  :Unsupported OS to crate virtual environment!")

    # Install the realtwin package in the virtual environment
    try:
        subprocess.check_call([python_executable, '-m', 'pip', 'install', pkg_name])
        print(f"  :Successfully installed {pkg_name} in the virtual environment.")
        return True
    except subprocess.CalledProcessError as e:
        print(f"  :Failed to install {pkg_name}. Error: {e}")
    except Exception as e:
        print(f"  :An unexpected error occurred: {e}")

    return False


def venv_delete(*, venv_name: str = "", venv_dir: str = "", verbose: bool = True) -> bool:
    """Delete the virtual environment in the specified folder with the specified name.

    Args:
        venv_name (str): the name of the virtual environment
        venv_dir (str): the path to the folder where the virtual environment will be deleted
        verbose (bool): whether to print the progress

    Example:
        >>> import realtwin as rt
        >>> rt.venv_delete(venv_name="my_venv", venv_dir="/path/to/dir")

    Returns:
        bool: True if the virtual environment is deleted successfully, False otherwise
    """

    # Default values for env_name and folder_path if not provided
    if not venv_name:
        venv_name = "venv_rt"

    if not venv_dir:
        venv_dir = os.getcwd()

    # Full path to the virtual environment
    venv_path = pf.path2linux(os.path.join(venv_dir, venv_name))

    # check if venv_path exists
    if not os.path.exists(venv_path):
        print(f"  :Virtual environment: {venv_path} does not exist, please check your inputs")
        return False

    if verbose:
        print(f"  :Deleting virtual environment '{venv_name}'")

    # Delete the virtual environment
    shutil.rmtree(venv_path)
    print(f"  :Virtual environment '{venv_name}' deleted successfully.")

    return True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

import requests
from datetime import datetime


def download_elevation_tif_by_bbox(bbox: tuple | list, output_file: str) -> bool:
    """Download elevation data (TIFF) from USGS National Map based on a bounding box.

    Args:
        bbox (tuple | list): A tuple or list containing the bounding box coordinates in the format (min_lon, min_lat, max_lon, max_lat).
        output_file (str): The path to save the downloaded TIFF file.

    Example:
        >>> from realtwin import download_elevation_tif_by_bbox
        >>> bbox = (-112.185, 36.056, -111.705, 36.368)  # Grand Canyon region
        >>> output_file = "elevation_data.tif"
        >>> download_elevation_tif_by_bbox(bbox, output_file)
        >>> # The function will download the elevation data and save it as "elevation_data.tif".

    Raises:
        ValueError: If the bounding box is not a tuple or list, or if it does not contain 4 coordinates.
        Exception: If the API request fails or if no data is available for the specified bounding box.
        Exception: If the download fails or if the file cannot be saved.

    Returns:
        bool: True if the download was successful, False otherwise.
    """

    # check input parameters
    if not isinstance(bbox, (tuple, list)):
        raise ValueError("Bounding box must be a tuple or list.")

    if len(bbox) != 4:
        raise ValueError("Bounding box must contain 4 coordinates (min_lon, min_lat, max_lon, max_lat).")

    # check whether the output file with .tif extension
    if not output_file.endswith(".tif"):
        raise ValueError("Output file must be a TIFF file.")

    # Download elevation data from USGS National Map
    # try:
    # USGS Elevation Data API endpoint
    usgs_api_url = "https://tnmaccess.nationalmap.gov/api/v1/products"

    # Create a bounding box WKT string
    # bbox_geom = box(*bbox)
    # bbox_wkt = bbox_geom.wkt

    # API parameters
    params = {
        "datasets": "National Elevation Dataset (NED) 1/3 arc-second",
        "bbox": ",".join([str(x) for x in bbox]),
        "outputFormat": "json",
        "extentType": "bbox"
    }

    # Make a request to the API
    response = requests.get(usgs_api_url, params=params, timeout=60 * 10) # 10 minutes timeout

    # Check for a successful response
    if response.status_code != 200:
        raise Exception(f"Failed to query USGS API: {response.status_code} {response.text}")

    # Parse the response
    data = response.json()
    if not data.get("items"):
        raise Exception("No data available for the specified bounding box.")

    # Download the first available GeoTIFF file
    # tiff_url = data["items"][0]["downloadURL"]
    tiff_url_list = {item["downloadURL"] for item in data["items"]}
    print("Downloaded URLs: ", tiff_url_list)

    # Extract date from each URL
    def extract_date(url):
        try:
            date_str = url.split('_')[-1].split('.')[0]
            return datetime.strptime(date_str, '%Y%m%d')
        except Exception:
            return None

    # Find the URL with the latest date
    latest_url = max(tiff_url_list, key=extract_date)
    print(f"Downloading GeoTIFF file from: {latest_url}")

    # Download the TIFF file
    tiff_response = requests.get(latest_url, stream=True, timeout=60 * 10)  # 10 minutes timeout
    if tiff_response.status_code == 200:
        # Get the total file size in MB
        total_size = int(tiff_response.headers.get('content-length', 0))
        total_size_mb = total_size / (1024 * 1024)
        print(f"Total file size: {total_size_mb:.2f} MB")

        downloaded_size = 0
        with open(output_file, "wb") as file:
            # 1 MB chunks
            for chunk in tiff_response.iter_content(chunk_size=1024 * 1024):
                if chunk:  # Filter out keep-alive chunks
                    file.write(chunk)
                    downloaded_size += len(chunk)
                    downloaded_mb = downloaded_size / (1024 * 1024)
                    print(f"Downloaded: {downloaded_mb:.2f} MB", end="\r")

        print(f"\nGeoTIFF file saved as: {output_file}")
    else:
        raise Exception(f"Failed to download GeoTIFF file: {tiff_response.status_code}")
    return None


if __name__ == "__main__":
    # Define bounding box (example: Grand Canyon region)
    bounding_box = (-112.185, 36.056, -111.705, 36.368)

    # Output file path
    output_path = "elevation_data.tif"

    # Download the elevation data
    download_elevation_tif_by_bbox(bounding_box, output_path)



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################


from pathlib import Path
from lxml import etree
import pandas as pd
import re
import subprocess


def name_without_suffixes(p: Path) -> str:
    """Return the name of the file without any suffixes."""
    # peel off each suffix until none remain
    while p.suffix:
        p = p.with_suffix('')
    return p.name


def parse_SUMO_ID(net_file: str) -> bool:
    """ Parse the SUMO network file to extract IDs (non internal) and update the net file.

    Args:
        net_file (str): Path to the SUMO network file.

    Returns:
        bool: True if the parsing and updating were successful, False otherwise.
    """

    try:
        # Parse the SUMO network file
        tree = etree.parse(net_file)
        root = tree.getroot()

        # Get all non-internal edge IDs
        sumoID_org = []
        sumoID_org.extend(
            edge.get("id")
            for edge in root.findall("edge")
            if edge.get("function") != "internal"
        )
        # Build replacement lists
        oldedgeID = []
        newedgeID = []

        for eid in sumoID_org:
            if not eid.startswith("-"):
                oldedgeID.append(eid)
                new_id = "-" + eid
                if new_id in sumoID_org:
                    new_id = "-" + new_id
                newedgeID.append(new_id)

        edgeID_replacement = pd.DataFrame({
            "oldedgeID": oldedgeID,
            "newedgeID": newedgeID
        })

        with open(net_file, "r", encoding="utf-8") as f:
            original_text = f.read()

        # Perform replacements in the file content
        for _, row in edgeID_replacement.iterrows():
            old_id = row["oldedgeID"]
            new_id = row["newedgeID"]
            original_text = re.sub(rf'(?<!index=)"{old_id}"', f'"{new_id}"', original_text)
            original_text = re.sub(f'lane id="{old_id}_', f'lane id="{new_id}_', original_text)
            original_text = re.sub(f'incLanes="{old_id}_', f'incLanes="{new_id}_', original_text)
            original_text = re.sub(f'intLanes="{old_id}_', f'intLanes="{new_id}_', original_text)
            original_text = re.sub(f' {old_id}_', f' {new_id}_', original_text)

        with open(net_file, "w", encoding="utf-8") as f:
            f.write(original_text)

        return True

    except Exception as e:
        print(f"Error parsing SUMO ID: {e}")
        return False


def parse_SUMO_to_OpenDrive(net_file: str) -> bool:
    """  Parse the SUMO network file to OpenDrive file."""
    file_dir = Path(net_file).parent
    file_name = name_without_suffixes(Path(net_file))
    output_file = file_dir / f"{file_name}.xodr"

    cmd = [
        "netconvert",
        "-s", f"{net_file}",
        "--opendrive-output", f"{output_file}",
        "--output.original-names", "true",
        "--junctions.scurve-stretch", "1.0"
    ]

    process = subprocess.Popen(cmd, shell=True)
    process.wait()

    if process.returncode != 0:
        print(f"Error converting SUMO to OpenDrive: {process.returncode}")
        return False
    return True


def match_SUMO_to_OpenDrive(xodr_file: str) -> bool:
    """ Match SUMO and OpenDrive IDs in the OpenDrive file."""

    try:

        with open(xodr_file, 'r', encoding='utf-8') as f:
            original_text = f.read()

        tree = etree.parse(xodr_file)
        root = tree.getroot()

        # Extract old/new edge IDs from <road>
        old_edge_ids = []
        new_edge_ids = []
        warning_printed = False
        for road in root.findall("road"):
            road_id = road.get("id")
            user_data = road.find("userData")
            if user_data is not None:
                sumo_id = user_data.get("value")

                if sumo_id.startswith("-"):
                    old_edge_ids.append(road_id)
                    new_id = f'{sumo_id[1:]}'
                    new_edge_ids.append(new_id)
                elif not warning_printed:
                    print("To ensure Real-Twin works properly, please make all SUMO edge ID start with '-'")
                    warning_printed = True

        # Extract old/new junction IDs from <junction>
        old_junction_ids = []
        new_junction_ids = []

        for junction in root.findall("junction"):
            junction_id = junction.get("id")
            junction_name = junction.get("name")
            old_junction_ids.append(junction_id)
            new_junction_ids.append(junction_name)

        edgeID = pd.DataFrame({'oldedgeID': old_edge_ids, 'newedgeID': new_edge_ids})
        junctionID = pd.DataFrame({'oldjunctionID': old_junction_ids, 'newjunctionID': new_junction_ids})

        # Apply edge ID replacements
        for _, row in edgeID.iterrows():
            old_id = row['oldedgeID']
            new_id = row['newedgeID']
            original_text = original_text.replace(f'id="{old_id}" junction=', f'id="{new_id}" junction=')
            original_text = original_text.replace(f'elementType="road" elementId="{old_id}"', f'elementType="road" elementId="{new_id}"')
            original_text = original_text.replace(f'incomingRoad="{old_id}"', f'incomingRoad="{new_id}"')
            original_text = original_text.replace(f'connectingRoad="{old_id}"', f'connectingRoad="{new_id}"')

        # Apply junction ID replacements
        for _, row in junctionID.iterrows():
            old_id = row['oldjunctionID']
            new_id = row['newjunctionID']
            original_text = original_text.replace(f'id="{old_id}">', f'id="{new_id}">')
            original_text = original_text.replace(f'elementType="junction" elementId="{old_id}"', f'elementType="junction" elementId="{new_id}"')
            original_text = original_text.replace(f'junction="{old_id}"', f'junction="{new_id}"')

        with open(xodr_file, 'w', encoding='utf-8') as f:
            f.write(original_text)

        return True

    except Exception as e:
        print(f"Error matching SUMO to OpenDrive: {e}")
        return False


def parse_SUMO_TLS_ID(path_matchup_table: str, path_net_file: str) -> bool:
    """Parse the SUMO network file to extract traffic light system IDs and update the net file.

    Args:
        path_matcup_table (str): Path to the SUMO matcup table file.
        path_net_file (str): Path to the SUMO network file.

    Returns:
        bool: True if the parsing and updating were successful, False otherwise.
    """
    try:
        # Read the matcup table
        MatchupTable_UserInput = pd.read_excel(path_matchup_table, skiprows=1, dtype=str)
        merged_columns = ["JunctionID_OpenDrive", "File_GridSmart", "Date_GridSmart",
                          "IntersectionName_GridSmart", "File_Synchro", "IntersectionID_Synchro", "Need calibration?"]
        MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill()

        signalized_junction = MatchupTable_UserInput[MatchupTable_UserInput['Turn_Synchro'].notna()]['JunctionID_OpenDrive'].unique()
        signalized_junction_string = ",".join(signalized_junction.astype(str))

        cmd = f'cmd /c "netconvert -s {path_net_file} --o {path_net_file} --tls.discard-loaded true --tls.set {signalized_junction_string}"'
        process = subprocess.Popen(cmd, shell=True)
        process.wait()

        return True

    except Exception as e:
        print(f"Error parsing SUMO TLS ID: {e}")
        return False



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################


class VissimPrep:
    """Class to prepare VISSIM simulation environment."""

    def __init__(self, **kwargs):
        """Initialize the VissimPrep class with optional parameters."""
        self.kwargs = kwargs
        self.verbose = kwargs.get('verbose', True)

    def prepare(self):
        """Prepare the VISSIM simulation environment."""
        if self.verbose:
            print("  :Preparing VISSIM simulation environment...")



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

from pathlib import Path
import shutil


def prepare_realtwin_configs(dest_dir: str = "") -> bool:
    """ Create a default/demo configuration file for the RealTwin package.

    Args:
        dest_dir (str): save the configuration file to destination dir.
            Defaults to "", which means the current working directory.

    Example:
        >>> import realtwin as rt
        >>> rt.create_configuration_file()  # will create a configuration file in the current working directory
        >>> rt.create_configuration_file(dest_dir="path/to/directory")  # will create in the specified directory

    Return:
        bool: True if the configuration file is created successfully.
    """
    # TDD

    path_config = Path(__file__).parent.parent / "data_lib/realtwin_config.yaml"

    if not dest_dir:
        dest_dir = Path.cwd()

    # check if the destination directory exists
    if not Path(dest_dir).exists():
        raise FileNotFoundError(f"Destination directory does not exist: {dest_dir}")

    # copy the configuration file to the destination directory
    dest_config = Path(dest_dir) / "realtwin_config.yaml"
    shutil.copy(path_config, dest_config)
    print(f"  :Configuration file created at {dest_config}")
    return True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################


def cali_aimsun(*, sel_algo: dict = None, input_config: dict = None, verbose: bool = True) -> bool:
    """Run AimSUN calibration based on the selected algorithm and input configuration.

    Note:
        This function is currently a placeholder and does not perform any actual calibration.
        It is intended to be implemented in the future.

    Args:
        sel_algo (dict): the dictionary of selected algorithm for turn_inflow and behavior. Defaults to None.
        input_config (dict): the dictionary contain configurations from input yaml file. Defaults to None.
        verbose (bool): print out processing message. Defaults to True.

    Returns:
        bool: True if calibration is successful, False otherwise.
    """

    print("  :Aimsun calibration is not implemented yet. Please check from future releases.")
    return False



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
import io
import re
import os
from pathlib import Path
import numpy as np
import pandas as pd
from realtwin.func_lib._c_abstract_scenario.rt_matchup_table_generation import generate_matchup_table


def process_signal_from_utdf(file_utdf: object) -> dict[str, pd.DataFrame]:
    """Process the signal data and return a dictionary of DataFrames.

    Args:
        file_utdf (object): the path to the UTDF file containing signal data.

    Returns:
        dict[str, pd.DataFrame]: A dictionary where keys are table names and values
        are DataFrames containing the data from those tables.
    """

    SignalDict = {}
    current_table = None
    current_table_data = []

    with open(file_utdf, 'r', encoding='utf-8') as f:
        file_lines = f.readlines()

    removal_flag = 0
    for line in file_lines:
        line = line.strip()

        # Check if it's a line to be skipped
        if removal_flag == 1:
            removal_flag = 0
            continue

        # Check if it's a table name (indicated by square brackets)
        if line.startswith("["):
            removal_flag = 1
            if current_table is None:
                # Remove square brackets and ]
                current_table = line[1:-1].split(',')[0].rstrip(']')
            else:
                # Store the previous table data in the dictionary
                if current_table_data:  # Check if there's data to store
                    # Read data as strings
                    df = pd.read_csv(io.StringIO(
                        '\n'.join(current_table_data)), dtype=str)
                    SignalDict[current_table] = df

                # Start a new table
                # Remove the square brackets and extra commas
                current_table = line[1:-1].split(',')[0].rstrip(']')
                current_table_data = []
        else:
            # Accumulate table data
            current_table_data.append(line)

    # Store the last table in the dictionary
    if current_table_data:
        SignalDict[current_table] = pd.read_csv(
            io.StringIO('\n'.join(current_table_data)), dtype=str)

    return SignalDict


# Helper function for sorting: extract the alphabetic base and numeric suffix
def sort_key(movement: str) -> tuple[int, int]:
    """Extracts the base and suffix from a movement string for sorting."""

    base_order = [
        'NBR', 'NBT', 'NBL', 'NER', 'NET', 'NEL',
        'EBR', 'EBT', 'EBL', 'SER', 'SET', 'SEL',
        'SBR', 'SBT', 'SBL', 'SWR', 'SWT', 'SWL',
        'WBR', 'WBT', 'WBL', 'NWR', 'NWT', 'NWL'
    ]
    match = re.match(r"([A-Z]+)(\d*)", movement)
    if match:
        base = match.group(1)
        suffix_str = match.group(2)
        suffix = int(suffix_str) if suffix_str else 0
        try:
            base_index = base_order.index(base)
        except ValueError:
            base_index = len(base_order)
        return (base_index, suffix)
    return (len(base_order), 0)


# Helper function: determine if a value is missing or equivalent to 0 (as a number or string).
def is_missing_or_zero(val) -> bool:
    """Check if a value is missing or equivalent to zero."""

    if pd.isna(val):
        return True
    val_str = str(val).strip()
    if val_str in {"", "0"}:
        return True

    try:
        if float(val_str) == 0:
            return True
    except ValueError:
        pass
    return False


def update_matchup_table(path_matchup_table: str, control_dir: str = "", traffic_dir: str = "") -> bool:
    """ Update the match table with data from user prepared demands and Synchro UTDF file.

    Args:
        path_matchup_table (str): the match table that contains the user input.
        control_dir (str): the directory where the Synchro UTDF files are located.
        traffic_dir (str): the directory where the traffic data files are located.
        Defaults to "", which means the current directory.

    Returns:
        bool: True if the matchup table is updated successfully, False otherwise.
    """

    # read the lookup table
    MatchupTable_UserInput = pd.read_excel(path_matchup_table, skiprows=1, dtype=str)

    # Forward fill missing values in merged columns
    merged_columns = ["JunctionID_OpenDrive", "File_Synchro"]
    MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill()
    MatchupTable_UserInput["Need calibration?"] = "Y"

    movement_order = ["NBR", "NBT", "NBL", "NBU", "EBR", "EBT", "EBL", "EBU",
                      "SBR", "SBT", "SBL", "SBU", "WBR", "WBT", "WBL", "WBU"]

    # Process each unique JunctionID_OpenDrive where File_GridSmart has input
    for junction_id in MatchupTable_UserInput["JunctionID_OpenDrive"].unique():
        subset = MatchupTable_UserInput[MatchupTable_UserInput["JunctionID_OpenDrive"] == junction_id]

        # Get the file path from File_GridSmart if available
        file_name = subset["File_GridSmart"].dropna().iloc[0] if not subset["File_GridSmart"].isna().all() else None

        if file_name:
            if '.' not in Path(file_name).name:
                file_name = file_name + '.xls'
            # Set Need calibration? to "N"
            MatchupTable_UserInput.loc[MatchupTable_UserInput["JunctionID_OpenDrive"]
                                       == junction_id, "Need calibration?"] = "N"

            # Load the GridSmart file
            gs_file_path = Path(traffic_dir) / file_name

            if not gs_file_path.suffix:   # suffix is empty if no extension
                gs_file_path = gs_file_path.with_suffix(".xls")   
                
            try:
                gs_data = pd.read_excel(gs_file_path, header=None, dtype=str)
            except FileNotFoundError:
                continue

            # Extract IntersectionName_GridSmart
            intersection_row = gs_data[gs_data.iloc[:, 0] == "Intersection"].index
            if not intersection_row.empty:
                intersection_col = gs_data.iloc[intersection_row[0], 1:].first_valid_index()
                if intersection_col is not None:
                    intersection_name = gs_data.iloc[intersection_row[0], intersection_col]
                    MatchupTable_UserInput.loc[MatchupTable_UserInput["JunctionID_OpenDrive"]
                                               == junction_id, "IntersectionName_GridSmart"] = intersection_name

            # Extract Date_GridSmart
            date_row = gs_data[gs_data.iloc[:, 0] == "Date"].index
            if not date_row.empty:
                date_col = gs_data.iloc[date_row[0], 1:].first_valid_index()
                if date_col is not None:
                    date_value = gs_data.iloc[date_row[0], date_col]
                    MatchupTable_UserInput.loc[MatchupTable_UserInput["JunctionID_OpenDrive"]
                                               == junction_id, "Date_GridSmart"] = date_value

            # Identify movement columns outside loop
            movement_columns = {}
            for col in gs_data.columns[1:]:
                for movement in ("Right", "Through", "Left"):
                    if gs_data.iloc[:, col].eq(movement).any():
                        movement_columns[movement] = col

            # Extract Turning Movements
            GridSmartTurns = []

            for direction, prefix in zip(["Northbound", "Eastbound", "Southbound", "Westbound"],
                                         ["NB", "EB", "SB", "WB"]):
                direction_row = gs_data[gs_data.iloc[:, 0] == direction].index
                if not direction_row.empty:
                    movement_row = direction_row[0]

                    # Check if movements exist and add to list
                    if "Right" in movement_columns and pd.notna(gs_data.iloc[movement_row,
                                                                             movement_columns["Right"]]):
                        GridSmartTurns.append(f"{prefix}R")
                    if "Through" in movement_columns and pd.notna(gs_data.iloc[movement_row,
                                                                               movement_columns["Through"]]):
                        GridSmartTurns.append(f"{prefix}T")
                    if "Left" in movement_columns and pd.notna(gs_data.iloc[movement_row,
                                                                            movement_columns["Left"]]):
                        GridSmartTurns.append(f"{prefix}L")
                        GridSmartTurns.append(f"{prefix}U")

            # Sort GridSmartTurns
            GridSmartTurns = sorted(
                GridSmartTurns,
                key=lambda x: movement_order.index(x) if x in movement_order else len(movement_order))

            # Fill Turn_GridSmart column
            rows_to_fill = subset.index.tolist()
            for i, turn in enumerate(GridSmartTurns):
                if i < len(rows_to_fill):
                    MatchupTable_UserInput.at[rows_to_fill[i], "Turn_GridSmart"] = turn
                else:
                    print(f'  :There are more turning movements in {file_name} '
                          f'than OpenDrive junction: {junction_id}.')

    # Initialize a cache for Synchro files to avoid re-reading files
    synchro_cache = {}

    # Process each unique JunctionID_OpenDrive for File_Synchro
    for junction_id in MatchupTable_UserInput["JunctionID_OpenDrive"].unique():
        subset = MatchupTable_UserInput[MatchupTable_UserInput["JunctionID_OpenDrive"] == junction_id]

        # Get the File_Synchro value (if provided)
        file_synchro_name = subset["File_Synchro"].dropna(
        ).iloc[0] if not subset["File_Synchro"].isna().all() else None
        if subset["IntersectionID_Synchro"].dropna().empty:
            continue

        intersection_id_synchro = subset["IntersectionID_Synchro"].dropna().iloc[0]
        # Use the cache if this file was already read
        if file_synchro_name in synchro_cache:
            signal_dict = synchro_cache[file_synchro_name]
        else:
            synchro_file_path = f"{os.path.join(control_dir, file_synchro_name)}"
            signal_dict = process_signal_from_utdf(synchro_file_path)
            synchro_cache[file_synchro_name] = signal_dict

        # Ensure the 'Lanes' table exists
        lanes_df = signal_dict.get('Lanes')
        if lanes_df is None:
            print(f'No "Lanes" table found in "{file_synchro_name}".')
            continue

        # 2.1: Subset rows where INTID equals IntersectionID_Synchro and RECORDNAME is in the allowed list.
        allowed_recordnames = ["Lanes", "Shared", "Phase1",
                               "PermPhase1", "Phase2", "PermPhase2", "Phase3", "PermPhase3"]
        subset_lanes = lanes_df[
            (lanes_df['INTID'].astype(str) == str(intersection_id_synchro)) & (
                lanes_df['RECORDNAME'].astype(str).isin(allowed_recordnames))
        ]

        for col in subset_lanes.columns:
            if not col.endswith('T'):
                continue

            base = col[:-1]  # Extract base like 'XY' from 'XYT'
            try:
                phaseid = subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Phase1', col].values[0]
                shareid_val = subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Shared', col].values[0]
            except IndexError:
                continue  # Skip if 'Phase1' or 'Shared' row not found
            except KeyError:
                continue  # Skip if col not in subset_lanes

            if pd.isna(phaseid) or pd.isna(shareid_val):
                continue  # Skip if any value is missing

            try:
                shareid = int(float(shareid_val))
                # print(shareid)
            except ValueError:
                continue

            if shareid == 0:
                continue

            if shareid in [1, 3] and base + 'L' in subset_lanes.columns:
                # Left turn logic
                l_col = base + 'L'
                l_phase = subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Phase1', l_col].values[0] if not subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Phase1', l_col].empty else np.nan
                l_perm = subset_lanes.loc[subset_lanes['RECORDNAME'] == 'PermPhase1', l_col].values[0] if not subset_lanes.loc[subset_lanes['RECORDNAME'] == 'PermPhase1', l_col].empty else np.nan
                if pd.isna(l_phase) and pd.isna(l_perm):
                    subset_lanes.loc[subset_lanes['RECORDNAME'] == 'PermPhase1', l_col] = phaseid

            if shareid in [2, 3] and base + 'R' in subset_lanes.columns:
                # Right turn logic
                r_col = base + 'R'
                r_phase = subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Phase1', r_col].values[0] if not subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Phase1', r_col].empty else np.nan
                r_perm = subset_lanes.loc[subset_lanes['RECORDNAME'] == 'PermPhase1', r_col].values[0] if not subset_lanes.loc[subset_lanes['RECORDNAME'] == 'PermPhase1', r_col].empty else np.nan
                if pd.isna(r_phase) and pd.isna(r_perm):
                    subset_lanes.loc[subset_lanes['RECORDNAME'] == 'Phase1', r_col] = phaseid

        if subset_lanes.empty:
            print(f'No matching records in Lanes for IntersectionID_Synchro '
                  f'{intersection_id_synchro} in file {file_synchro_name}.')
            continue

        # 2.2: Reindex the row numbering (reset the index)
        subset_lanes.reset_index(drop=True, inplace=True)

        # Drop columns based on the condition applied to row 1 (first row, index 0)
        cols_to_drop = []
        for col in subset_lanes.columns:
            # Get the value in the first row (row 1)
            val_first = subset_lanes.at[0, col]
            if is_missing_or_zero(val_first):
                # Exception 1: if any value from row 3 onward (i.e. index 2 and beyond) is valid, we keep the column.
                if subset_lanes.shape[0] > 2:
                    subsequent_valid = subset_lanes[col].iloc[2:].apply(
                        lambda x: not is_missing_or_zero(x)).any()
                else:
                    subsequent_valid = False

                # Exception 2: for any column ending with 'R', check if the corresponding XYT column meets the criteria.
                exception2_keep = False
                if col.endswith("R"):
                    col_t = col[:-1] + "T"
                    if col_t in subset_lanes.columns and subset_lanes.shape[0] > 1:
                        try:
                            val_first_t = float(subset_lanes.at[0, col_t])
                            val_second_t = float(
                                subset_lanes.at[1, col_t]) if subset_lanes.shape[0] > 1 else 0
                            if val_first_t > 0 and val_second_t > 1:
                                exception2_keep = True
                        except ValueError:
                            pass

                # Drop the column only if neither exception applies.
                if not subsequent_valid and not exception2_keep:
                    cols_to_drop.append(col)

        # Drop the columns that do not meet the exceptions
        subset_lanes = subset_lanes.copy()
        subset_lanes.drop(columns=cols_to_drop, inplace=True)

        # Finally, get the column names and save them in the variable 'movements',
        # while excluding the unwanted columns
        movements = [col for col in subset_lanes.columns if col not in [
            'RECORDNAME', 'INTID', 'PED', 'HOLD']]

        # Sort the original movements using the sort_key
        sorted_movements = sorted(movements, key=sort_key)

        # For each sorted movement that ends with 'L', immediately add the variant (replace ending 'L' with 'U')
        enhanced_movements = []
        for mov in sorted_movements:
            enhanced_movements.append(mov)
            if mov.endswith("L"):
                variant = mov[:-1] + "U"
                enhanced_movements.append(variant)

        # Fill the Turn_Synchro column for the current JunctionID_OpenDrive
        rows_to_fill = subset.index.tolist()
        for i, movement in enumerate(enhanced_movements):
            if i < len(rows_to_fill):
                MatchupTable_UserInput.at[rows_to_fill[i], "Turn_Synchro"] = movement
            else:
                print(f'  :There are more turning movements in {file_synchro_name} '
                      f'than OpenDrive junction {junction_id}.')

    # save the updated MatchupTable_UserInput to the same file
    generate_matchup_table(MatchupTable_UserInput, path_matchup_table)
    return MatchupTable_UserInput


def generate_turn_demand(*, path_matchup_table: str,
                         control_dir: str,
                         traffic_dir: str, output_dir: str = "") -> list[pd.DataFrame]:
    """ Generate turn demand from user input lookup table and Synchro UTDF files.

    Args:
        path_matchup_table (str): Path to the matchup table with user input.
        control_dir (str): Directory where Synchro UTDF files are located.
            Defaults to "", which means the current directory.
        output_dir (str): Directory to save the output files.
            Defaults to "", which means the current directory.
        traffic_dir (str): Directory where demand files are located.
            Defaults to "Traffic".

    See Also:
        demand_dir: check sample demand files in datasets/Traffic directory

    Example:
        >>> path_matchup_table = "./MatchupTable_OpenDrive_with user input.xlsx"
        >>> TurnDf, IDRef = generate_turn_demand(path_matchup_table, signal_dir="",
            output_dir="./Output", demand_dir="Traffic")

    Returns:
        list[pd.DataFrame]: A list containing two DataFrames:
            - TurnDf: DataFrame with turn demand data.
            - IDRef: DataFrame with reference IDs for OpenDrive turns (demand lookup table).
    """

    if isinstance(path_matchup_table, str):
        # get the updated lookup table
        try:
            MatchupTable_UserInput = pd.read_excel(path_matchup_table, skiprows=1, dtype=str)
            # MatchupTable_UserInput = update_matchup_table(path_matchup_table, control_dir, traffic_dir)
        except Exception as e:
            raise Exception("Error loading user updated lookup table") from e
    elif isinstance(path_matchup_table, pd.DataFrame):
        # Use the provided DataFrame directly
        MatchupTable_UserInput = path_matchup_table
    else:
        raise ValueError("path_matchup_table must be a string or a DataFrame.")

    merged_columns = ["JunctionID_OpenDrive", "File_Synchro"]
    MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill()

    turn_values = ["NBR", "NBT", "NBL", "NBU", "EBR", "EBT", "EBL", "EBU",
                   "SBR", "SBT", "SBL", "SBU", "WBR", "WBT", "WBL", "WBU"]
    TurnDf_list = []
    IDRef_list = []

    # Process each unique JunctionID_OpenDrive where File_GridSmart has input
    for junction_id in MatchupTable_UserInput["JunctionID_OpenDrive"].unique():
        subset = MatchupTable_UserInput[MatchupTable_UserInput["JunctionID_OpenDrive"] == junction_id]

        # Get the file path from File_GridSmart if available
        file_name = subset["File_GridSmart"].dropna().iloc[0] if not subset["File_GridSmart"].isna().all() else None

        if file_name:
            if '.' not in Path(file_name).name:
                file_name = file_name + '.xls'            
            # Retrieve the IntersectionName_GridSmart
            intersection_name = subset["IntersectionName_GridSmart"].dropna().iloc[0] if not subset["IntersectionName_GridSmart"].isna().all() else "Unknown"
            # Create df_lookup with predefined Turn values
            df_lookup = pd.DataFrame({"Turn": turn_values})
            # Assign IntersectionName
            df_lookup["IntersectionName"] = intersection_name
            # Initialize OpenDriveFromID and OpenDriveToID as empty strings
            df_lookup["OpenDriveFromID"] = ""
            df_lookup["OpenDriveToID"] = ""
            # Map values from subset based on Turn_GridSmart
            for idx, row in df_lookup.iterrows():
                turn = row["Turn"]
                match = subset[subset["Turn_GridSmart"] == turn]
                if not match.empty:
                    df_lookup.at[idx, "OpenDriveFromID"] = match["FromRoadID_OpenDrive"].values[0] if not match["FromRoadID_OpenDrive"].isna().all() else ""
                    df_lookup.at[idx, "OpenDriveToID"] = match["ToRoadID_OpenDrive"].values[0] if not match["ToRoadID_OpenDrive"].isna().all() else ""

            IDRef_list.append(df_lookup)

        # Get the file path from File_GridSmart if available
        file_name = subset["File_GridSmart"].dropna().iloc[0] if not subset["File_GridSmart"].isna().all() else None

        if file_name:
            if '.' not in Path(file_name).name:
                file_name = file_name + '.xls'
            gs_file_path = os.path.join(traffic_dir, file_name)
            # gs_file_path = f"RealTwinDemand/{file_name}"

            # Check if the file exists before processing
            if os.path.exists(gs_file_path):
                # Load the Excel file
                df = pd.read_excel(gs_file_path, header=None)  # Read without predefined headers

                # Find the first row that contains a time value in the first column
                time_row_index = df[df[0].astype(str).str.match(r'^\d{1,2}:\d{2}$', na=False)].index.min()

                # Determine the starting row for data extraction
                start_row = time_row_index - 2 if pd.notna(time_row_index) else None

                if start_row is not None:
                    # Read the file again with proper headers starting from the determined row
                    df_data = pd.read_excel(gs_file_path, header=[start_row, start_row + 1])

                    # Fill merged cells in the first row
                    df_data.columns = df_data.columns.to_frame().ffill().agg("".join, axis=1)

                    # Remove spaces from column names
                    df_data.columns = [col.replace(" ", "") for col in df_data.columns]

                    # Rename the first column to "Time"
                    df_data.rename(columns={df_data.columns[0]: "Time"}, inplace=True)

                    # Drop fully empty columns
                    df_data.dropna(axis=1, how='all', inplace=True)

                    # Drop data of "Total"
                    df_data = df_data[df_data["Time"] != "Total"]

                    # Convert data columns to numeric
                    for col in df_data.columns[1:]:  # Excluding 'Time' column
                        df_data[col] = pd.to_numeric(df_data[col], errors='coerce').fillna(0).astype(int)

                    # Remove 'Unassigned' columns
                    df_data = df_data.loc[:, ~df_data.columns.str.contains(r'Unassigned', na=False)]

                    # Standardize column names for directions
                    df_data.columns = [col.replace("Northbound", "NB")
                                       .replace("Southbound", "SB")
                                       .replace("Westbound", "WB")
                                       .replace("Eastbound", "EB") for col in df_data.columns]

                    # Define expected columns and ensure all are present
                    expected_columns = ["IntersectionName", "Time", "NBR", "NBT", "NBL", "NBU",
                                        "EBR", "EBT", "EBL", "EBU", "SBR", "SBT", "SBL", "SBU",
                                        "WBR", "WBT", "WBL", "WBU"]
                    df_data = df_data.reindex(columns=expected_columns, fill_value="")
                    # df_data = df_data.reindex(columns=expected_columns, fill_value=None)

                    # Fill IntersectionName using IntersectionName_GridSmart from MatchupTable_UserInput
                    if not subset["IntersectionName_GridSmart"].isna().all():
                        intersection_name = subset["IntersectionName_GridSmart"].dropna().iloc[0]
                    else:
                        intersection_name = "Unknown"
                    df_data["IntersectionName"] = intersection_name

                    # Append processed data to list
                    TurnDf_list.append(df_data)

    TurnDf = pd.concat(TurnDf_list, ignore_index=True) if TurnDf_list else pd.DataFrame()

    IDRef = pd.concat(IDRef_list, ignore_index=True) if IDRef_list else pd.DataFrame()
    IDRef = IDRef[["IntersectionName", "Turn", "OpenDriveFromID", "OpenDriveToID"]]

    if output_dir:
        # Ensure the output directory exists
        os.makedirs(output_dir, exist_ok=True)
        # Save the DataFrame to a CSV file
        TurnDf.to_excel(os.path.join(output_dir, "GridSmart_demand.xlsx"), index=False)
        IDRef.to_excel(os.path.join(output_dir, "GridSmart_lookuptable.xlsx"), index=False)

    # replace "" to numpy.nan
    # TurnDf = TurnDf.replace("", np.nan)
    # IDRef = IDRef.replace("", np.nan)
    IDRef = IDRef.dropna(subset=['OpenDriveFromID', 'OpenDriveToID'])

    # drop '' in column OpenDriveFromID and OpenDriveToID
    IDRef = IDRef[IDRef["OpenDriveFromID"].astype(str) != ""]
    IDRef = IDRef[IDRef["OpenDriveToID"].astype(str) != ""]
    return [TurnDf, IDRef]

# if __name__ == "__main__":
#     path_matchup_table = "./MatchupTable_OpenDrive_with user input.xlsx"
#     output_dir = "./"
#     TurnDf, IDRef = generate_turn_demand(path_matchup_table, output_dir="./")



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""Control of module imports from func_lib."""

# _a_install_simulator: Functions to install and check simulation environments
from realtwin.func_lib._a_install_simulator.check_sim_env import (is_sumo_installed,
                                                                  is_vissim_installed,
                                                                  is_aimsun_installed)
from realtwin.func_lib._a_install_simulator.inst_sumo import (install_sumo,
                                                              install_sumo_windows,
                                                              install_sumo_linux,
                                                              install_sumo_macos)

# _b_load_inputs: Functions to load input configurations and data
from realtwin.func_lib._b_load_inputs.loader_config import load_input_configs


# _c_abstract_scenario: Abstract classes and functions for scenarios
from realtwin.func_lib._c_abstract_scenario._abstractScenario import (AbstractScenario,
                                                                      load_traffic_volume,
                                                                      load_traffic_turning_ratio,
                                                                      load_control_signal)

from realtwin.func_lib._c_abstract_scenario._network import Network, OpenDriveNetwork, OSMRoad
from realtwin.func_lib._c_abstract_scenario._traffic import Traffic
from realtwin.func_lib._c_abstract_scenario._control import Control
from realtwin.func_lib._c_abstract_scenario._application import Application

# _d_concrete_scenario: Concrete implementations of scenarios
from realtwin.func_lib._d_concrete_scenario._concreteScenario import ConcreteScenario
from realtwin.func_lib._d_concrete_scenario._supply import Supply
from realtwin.func_lib._d_concrete_scenario._demand import Demand
from realtwin.func_lib._d_concrete_scenario._behavior import Behavior
from realtwin.func_lib._d_concrete_scenario._route import Route
from realtwin.func_lib._d_concrete_scenario._trafficControl import TrafficControl

# _e_simulation: Functions to prepare and run simulations
from realtwin.func_lib._e_simulation._generate_simulation import SimPrep
from realtwin.func_lib._e_simulation._sumo import SUMOPrep
from realtwin.func_lib._e_simulation._aimsun import AimsunPrep
from realtwin.func_lib._e_simulation._vissim import VissimPrep

# _f_calibration: Calibration functions and algorithms for different simulators

from realtwin.func_lib._f_calibration.algo_sumo.cali_behavior import BehaviorCali
from realtwin.func_lib._f_calibration.algo_sumo.cali_turn_inflow import TurnInflowCali

from realtwin.func_lib._f_calibration.calibration_sumo import cali_sumo
from realtwin.func_lib._f_calibration.calibration_aimsun import cali_aimsun
from realtwin.func_lib._f_calibration.calibration_vissim import cali_vissim


__all__ = [
    # _a_install_simulator
    "is_sumo_installed",
    "is_vissim_installed",
    "is_aimsun_installed",
    "install_sumo",
    "install_sumo_windows",
    "install_sumo_linux",
    "install_sumo_macos",

    # _b_load_inputs
    "load_input_configs",

    # _c_abstract_scenario
    "AbstractScenario",
    "load_traffic_volume",  # Load traffic volume data
    "load_traffic_turning_ratio",  # Load traffic turning ratio data
    "load_control_signal",  # Load control signal data
    # "Network",  # Network class for road networks
    "OpenDriveNetwork",  # OpenDrive network class for OpenDrive format
    "OSMRoad",  # OSMRoad class for OpenStreetMap road representation
    # "Traffic",
    # "Control",
    # "Application",

    # _d_concrete_scenario
    "ConcreteScenario",
    # "Supply",  # Supply class for concrete supply scenarios
    # "Demand",  # Demand class for concrete demand scenarios
    # "Behavior",  # Behavior class for concrete behavior scenarios
    # "Route",  # Route class for concrete route scenarios
    # "TrafficControl",  # TrafficControl class for concrete traffic control scenarios

    # _e_simulation
    "SimPrep",  # Simulation preparation class
    "SUMOPrep",  # SUMO simulation preparation class
    "AimsunPrep",  # Aimsun simulation preparation class
    "VissimPrep",  # Vissim simulation preparation class

    # _f_calibration
    "cali_sumo",  # Calibration function for SUMO simulator
    "cali_aimsun",  # Calibration function for Aimsun simulator
    "cali_vissim",  # Calibration function for Vissim simulator
    "BehaviorCali",  # Behavior calibration algorithm for SUMO simulator
    "TurnInflowCali"  # Turn inflow calibration algorithm for SUMO simulator

    # _g_analyzer

    # _h_visualization
]
